from __future__ import print_function
"""
Test for basic virt-who operations.

Copyright (C) 2014 Radek Novacek <rnovacek@redhat.com>

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""

import sys
import copy
import os
import pytest
import six

from mock import patch, Mock, call

from base import TestBase

from virtwho import util
from virtwho.config import VW_GLOBAL, VW_ENV_CLI_SECTION_NAME
from virtwho.parser import parse_options, OptionError
from virtwho.executor import Executor


class TestOptions(TestBase):
    NO_GENERAL_CONF = {'global': {}}

    def setUp(self):
        self.clearEnv()

    def tearDown(self):
        self.clearEnv()
        super(TestBase, self).tearDown()

    def clearEnv(self):
        for key in list(os.environ.keys()):
            if key.startswith("VIRTWHO"):
                del os.environ[key]

    def setUpParseFile(self, parseFileMock):
        parseFileMock.return_value = TestOptions.NO_GENERAL_CONF

    @patch('virtwho.log.getLogger')
    @patch('virtwho.config.parse_file')
    def test_default_cmdline_options(self, parseFile, getLogger):
        self.setUpParseFile(parseFile)
        sys.argv = ["virtwho.py"]
        _, options = parse_options()
        self.assertFalse(options[VW_GLOBAL]['debug'])
        self.assertFalse(options[VW_GLOBAL]['oneshot'])
        self.assertEqual(options[VW_GLOBAL]['interval'], 3600)
        self.assertEqual(options[VW_GLOBAL]['reporter_id'], util.generate_reporter_id())

    @patch('virtwho.log.getLogger')
    @patch('virtwho.config.parse_file')
    def test_minimum_interval_options(self, parseFile, getLogger):
        self.setUpParseFile(parseFile)
        sys.argv = ["virtwho.py", "--interval=5"]
        _, options = parse_options()
        self.assertEqual(options[VW_GLOBAL]['interval'], 3600)

        sys.argv = ["virtwho.py"]
        os.environ["VIRTWHO_INTERVAL"] = '1'

        _, options = parse_options()
        self.assertEqual(options[VW_GLOBAL]['interval'], 3600)

        self.clearEnv()
        bad_conf = {'global': {'interval': '1'}}
        parseFile.return_value = bad_conf

        _, options = parse_options()
        self.assertEqual(options[VW_GLOBAL]['interval'], 3600)

    @patch('virtwho.log.getLogger')
    @patch('virtwho.config.parse_file')
    def test_empty_interval_options(self, parseFile, getLogger):
        self.setUpParseFile(parseFile)
        sys.argv = ["virtwho.py", "--interval="]
        _, options = parse_options()
        self.assertEqual(options[VW_GLOBAL]['interval'], 3600)

        sys.argv = ["virtwho.py"]
        os.environ["VIRTWHO_INTERVAL"] = ''

        _, options = parse_options()
        self.assertEqual(options[VW_GLOBAL]['interval'], 3600)

        self.clearEnv()
        bad_conf = {'global': {'interval': ''}}
        parseFile.return_value = bad_conf

        _, options = parse_options()
        self.assertEqual(options[VW_GLOBAL]['interval'], 3600)

    @pytest.mark.skipif(not six.PY2, reason="test only runs with python 2 virt-who")
    @patch('virtwho.log.getLogger')
    @patch('virtwho.config.parse_file')
    def test_options_consistency(self, parseFile, getLogger):
        self.setUpParseFile(parseFile)
        sys.argv = ["virtwho.py", "--libvirt", "--esx-username=admin"]
        self.assertRaises(OptionError, parse_options)

    @pytest.mark.skipif(not six.PY2, reason="test only runs with python 2 virt-who")
    @patch('virtwho.log.getLogger')
    @patch('virtwho.config.parse_file')
    def test_options_consistency_reverse_order(self, parseFile, getLogger):
        self.setUpParseFile(parseFile)
        sys.argv = ["virtwho.py", "--esx-username=admin", "--libvirt"]
        self.assertRaises(OptionError, parse_options)

    @pytest.mark.skipif(not six.PY2, reason="test only runs with python 2 virt-who")
    @patch('virtwho.log.getLogger')
    @patch('virtwho.config.parse_file')
    def test_options_missing_virt_backend(self, parseFile, getLogger):
        self.setUpParseFile(parseFile)
        sys.argv = ["virtwho.py", "--sam", "--esx-username=admin"]
        self.assertRaises(OptionError, parse_options)

    @pytest.mark.skipif(not six.PY2, reason="test only runs with python 2 virt-who")
    @patch('virtwho.log.getLogger')
    @patch('virtwho.config.parse_file')
    def test_options_order(self, parseFile, getLogger):
        self.setUpParseFile(parseFile)
        sys.argv = ["virtwho.py", "--libvirt-username=admin", "--libvirt"]
        _, options = parse_options()
        self.assertEqual(options[VW_ENV_CLI_SECTION_NAME]['type'], "libvirt")
        self.assertEqual(options[VW_ENV_CLI_SECTION_NAME]['username'], "admin")

    @pytest.mark.skipif(not six.PY2, reason="test only runs with python 2 virt-who")
    @patch('virtwho.log.getLogger')
    @patch('virtwho.config.parse_file')
    def test_options_hierarchy_for_reporter_id(self, parseFile, getLogger):
        # Set the value in all three possible locations
        # Mock /etc/virt-who.conf file
        global_conf_dict = {
            'global': {
                'reporter_id': "/etc/virt-who.conf"
            }
        }
        parseFile.side_effect = lambda x: copy.deepcopy(global_conf_dict)
        # cli option
        sys.argv = ["virtwho.py", "--reporter-id=cli"]
        # environment var
        os.environ["VIRTWHO_REPORTER_ID"] = "env"
        _, options = parse_options()
        # cli option should beat environment vars and virt-who.conf
        self.assertEqual(options[VW_GLOBAL]['reporter_id'], "cli")

        sys.argv = ["virtwho.py"]

        _, options = parse_options()
        self.assertEqual(options[VW_GLOBAL]['reporter_id'], "env")

        self.clearEnv()

        _, options = parse_options()
        self.assertEqual(options[VW_GLOBAL]['reporter_id'], "/etc/virt-who.conf")

        parseFile.side_effect = lambda x: {}

        _, options = parse_options()
        self.assertEqual(options[VW_GLOBAL]['reporter_id'], util.generate_reporter_id())

    @patch('virtwho.log.getLogger')
    @patch('virtwho.config.parse_file')
    def test_options_debug(self, parseFile, getLogger):
        self.setUpParseFile(parseFile)
        sys.argv = ["virtwho.py", "-d"]
        _, options = parse_options()
        self.assertTrue(options[VW_GLOBAL]['debug'])

    @pytest.mark.skipif(not six.PY2, reason="test only runs with python 2 virt-who")
    @patch('virtwho.log.getLogger')
    @patch('virtwho.config.parse_file')
    def test_options_virt(self, parseFile, getLogger):
        self.setUpParseFile(parseFile)
        for virt in ['esx', 'hyperv', 'rhevm']:
            self.clearEnv()
            sys.argv = ["virtwho.py", "--%s" % virt, "--%s-owner=owner" % virt,
                        "--%s-server=localhost" % virt,
                        "--%s-username=username" % virt,
                        "--%s-password=password" % virt]
            _, options = parse_options()
            self.assertEqual(options[VW_ENV_CLI_SECTION_NAME]['type'], virt)
            self.assertEqual(options[VW_ENV_CLI_SECTION_NAME]['owner'], 'owner')
            if virt == 'esx':
                self.assertEqual(options[VW_ENV_CLI_SECTION_NAME]['server'], 'https://localhost')
            elif virt == 'rhevm':
                self.assertEqual(options[VW_ENV_CLI_SECTION_NAME]['server'], 'https://localhost:8443/')
            else:
                self.assertEqual(options[VW_ENV_CLI_SECTION_NAME]['server'], 'localhost')
            self.assertEqual(options[VW_ENV_CLI_SECTION_NAME]['username'], 'username')
            self.assertEqual(options[VW_ENV_CLI_SECTION_NAME]['password'], 'password')

            sys.argv = ["virtwho.py"]
            virt_up = virt.upper()
            os.environ["VIRTWHO_%s" % virt_up] = "1"
            os.environ["VIRTWHO_%s_OWNER" % virt_up] = "xowner"
            os.environ["VIRTWHO_%s_ENV" % virt_up] = "xenv"
            os.environ["VIRTWHO_%s_SERVER" % virt_up] = "xlocalhost"
            os.environ["VIRTWHO_%s_USERNAME" % virt_up] = "xusername"
            os.environ["VIRTWHO_%s_PASSWORD" % virt_up] = "xpassword"
            _, options = parse_options()
            self.assertEqual(options[VW_ENV_CLI_SECTION_NAME]['type'], virt)
            self.assertEqual(options[VW_ENV_CLI_SECTION_NAME]['owner'], 'xowner')
            if virt == 'esx':
                self.assertEqual(options[VW_ENV_CLI_SECTION_NAME]['server'], 'https://xlocalhost')
            elif virt == 'rhevm':
                self.assertEqual(options[VW_ENV_CLI_SECTION_NAME]['server'], 'https://xlocalhost:8443/')
            else:
                self.assertEqual(options[VW_ENV_CLI_SECTION_NAME]['server'], 'xlocalhost')
            self.assertEqual(options[VW_ENV_CLI_SECTION_NAME]['username'], 'xusername')
            self.assertEqual(options[VW_ENV_CLI_SECTION_NAME]['password'], 'xpassword')

    @pytest.mark.skipif(not six.PY2, reason="test only runs with python 2 virt-who")
    @patch('virtwho.log.getLogger')
    @patch('virtwho.config.parse_file')
    def test_options_virt_satellite(self, parse_file, getLogger):
        self.setUpParseFile(parse_file)
        for virt in ['hyperv', 'esx', 'rhevm']:
            self.clearEnv()
            sys.argv = ["virtwho.py",
                        "--satellite",
                        "--satellite-server=localhost",
                        "--satellite-username=username",
                        "--satellite-password=password",
                        "--%s" % virt,
                        "--%s-server=localhost" % virt,
                        "--%s-username=username" % virt,
                        "--%s-password=password" % virt]
            _, options = parse_options()
            self.assertEqual(options[VW_ENV_CLI_SECTION_NAME]['type'], virt)
            if virt == 'rhevm':
                self.assertEqual(options[VW_ENV_CLI_SECTION_NAME]['server'], 'https://localhost:8443/')
            elif virt == 'esx':
                self.assertEqual(options[VW_ENV_CLI_SECTION_NAME]['server'], 'https://localhost')
            else:
                self.assertEqual(options[VW_ENV_CLI_SECTION_NAME]['server'], 'localhost')
            self.assertEqual(options[VW_ENV_CLI_SECTION_NAME]['username'], 'username')
            self.assertEqual(options[VW_ENV_CLI_SECTION_NAME]['password'], 'password')

            sys.argv = ["virtwho.py"]
            virt_up = virt.upper()
            os.environ["VIRTWHO_SATELLITE"] = "1"
            os.environ["VIRTWHO_SATELLITE_SERVER"] = "xlocalhost"
            os.environ["VIRTWHO_SATELLITE_USERNAME"] = "xusername"
            os.environ["VIRTWHO_SATELLITE_PASSWORD"] = "xpassword"
            os.environ["VIRTWHO_%s" % virt_up] = "1"
            os.environ["VIRTWHO_%s_OWNER" % virt_up] = 'xowner'
            os.environ["VIRTWHO_%s_ENV" % virt_up] = 'xenv'
            os.environ["VIRTWHO_%s_SERVER" % virt_up] = "xlocalhost"
            os.environ["VIRTWHO_%s_USERNAME" % virt_up] = "xusername"
            os.environ["VIRTWHO_%s_PASSWORD" % virt_up] = "xpassword"
            _, options = parse_options()
            self.assertEqual(options[VW_ENV_CLI_SECTION_NAME]['type'], virt)
            if virt == 'rhevm':
                self.assertEqual(options[VW_ENV_CLI_SECTION_NAME]['server'], 'https://xlocalhost:8443/')
            elif virt == 'esx':
                self.assertEqual(options[VW_ENV_CLI_SECTION_NAME]['server'], 'https://xlocalhost')
            else:
                self.assertEqual(options[VW_ENV_CLI_SECTION_NAME]['server'], 'xlocalhost')
            self.assertEqual(options[VW_ENV_CLI_SECTION_NAME]['owner'], 'xowner')
            self.assertEqual(options[VW_ENV_CLI_SECTION_NAME]['username'], 'xusername')
            self.assertEqual(options[VW_ENV_CLI_SECTION_NAME]['password'], 'xpassword')

    @pytest.mark.skipif(not six.PY2, reason="test only runs with python 2 virt-who")
    @patch('virtwho.log.getLogger')
    @patch('virtwho.config.parse_file')
    def test_missing_option(self, parseFile, getLogger):
        self.setUpParseFile(parseFile)
        for smType in ['satellite', 'sam']:
            for virt in ['libvirt', 'esx', 'hyperv', 'rhevm', 'kubevirt']:
                for missing in ['server', 'username', 'password', 'owner']:
                    self.clearEnv()
                    sys.argv = ["virtwho.py", "--%s" % smType, "--%s" % virt]
                    if virt in ['libvirt', 'esx', 'hyperv', 'rhevm']:
                        if missing != 'server':
                            sys.argv.append("--%s-server=localhost" % virt)
                        if missing != 'username':
                            sys.argv.append("--%s-username=username" % virt)
                        if missing != 'password':
                            sys.argv.append("--%s-password=password" % virt)
                        if missing != 'owner':
                            sys.argv.append("--%s-owner=owner" % virt)

                    if virt not in ('libvirt', 'kubevirt') and missing != 'password':
                        if smType == 'satellite' and missing in ['owner']:
                            continue
                        self.assertRaises(OptionError, parse_options)


class TestExecutor(TestBase):

    @patch.object(Executor, 'terminate_threads')
    @patch('virtwho.executor.time')
    def test_wait_on_threads(self, mock_time, mock_terminate_threads):
        """
        Tests that, given no kwargs, the wait_on_threads method will wait until
        all threads is_terminated method returns True.
        Please note that a possible consequence of something going wrong in
        the wait on threads method (with no kwargs) could cause this test to
        never quit.
        """
        # Create a few mock threads
        # The both will return False the first time is_terminated is called
        # Only the second mock thread will wait not return True until the
        # third call of is_terminated
        mock_thread1 = Mock()
        mock_thread1.is_terminated = Mock(side_effect=[False, True])
        mock_thread2 = Mock()
        mock_thread2.is_terminated = Mock(side_effect=[False, False, True])

        threads = [mock_thread1, mock_thread2]

        mock_time.sleep = Mock()
        Executor.wait_on_threads(threads)
        mock_time.sleep.assert_has_calls([
            call(1),
            call(1),
        ])
        mock_terminate_threads.assert_not_called()

    def test_terminate_threads(self):
        threads = [Mock(), Mock()]
        Executor.terminate_threads(threads)
        for mock_thread in threads:
            mock_thread.stop.assert_called()
            mock_thread.join.assert_called()



from __future__ import print_function

#
# Module for abstraction of all virtualization backends, part of virt-who
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#

from fake_server import FakeServer

from multiprocessing import Value


class FakeVirt(FakeServer):

    virt_type = None

    """
    Base class for fake virt backends like RHEVM or ESX
    """
    def __init__(self, handler_class, host='localhost', port=None):
        super(FakeVirt, self).__init__(handler_class, host, port)
        self._data_version = Value('d', 0)

    @property
    def username(self):
        return 'A!bc3#\'"'

    @property
    def password(self):
        return 'A!bc3#\'"'

    @property
    def data_version(self):
        return self._data_version.value

    @data_version.setter
    def data_version(self, version):
        self._data_version.value = version

    def virt_who_config(self):
        """
        This method return configuration file that can be used by virt-who for testing
        :return: string with configuration file
        """
        conf_file_content = """
[test]
type=%s
server=http://localhost:%s
username=%s
password=%s
owner=admin
        """ % (self.virt_type, self.port, self.username, self.password)
        return conf_file_content



from __future__ import print_function
import os
import sys
import shutil
import tempfile
import pytest
import six

from mock import patch, Mock, DEFAULT, MagicMock, ANY

from base import TestBase, unittest

from virtwho.config import VirtConfigSection, DestinationToSourceMapper, VW_ENV_CLI_SECTION_NAME,\
    init_config
from virtwho.manager import Manager
from virtwho.manager.subscriptionmanager import SubscriptionManager
from virtwho.virt import Guest, Hypervisor, HostGuestAssociationReport, DomainListReport, AbstractVirtReport, StatusReport
from virtwho.parser import parse_options


xvirt = type("", (), {'CONFIG_TYPE': 'xxx'})()


class TestSubscriptionManager(TestBase):
    guestList = [
        Guest('222', xvirt.CONFIG_TYPE, Guest.STATE_RUNNING),
        Guest('111', xvirt.CONFIG_TYPE, Guest.STATE_RUNNING),
        Guest('333', xvirt.CONFIG_TYPE, Guest.STATE_RUNNING),
    ]
    mapping = {
        'hypervisors': [Hypervisor('123', guestList, name='TEST_HYPERVISOR'),
                        Hypervisor('123', guestList, name='TEST_HYPERVISOR2')]
    }
    hypervisor_id = "HYPERVISOR_ID"
    uep_connection = None

    @classmethod
    @patch('rhsm.config.initConfig')
    @patch('rhsm.certificate.create_from_file')
    def setUpClass(cls, rhsmcert, rhsmconfig):
        super(TestSubscriptionManager, cls).setUpClass()
        config = VirtConfigSection.from_dict({'type': 'libvirt'}, 'test', None)
        cls.tempdir = tempfile.mkdtemp()
        with open(os.path.join(cls.tempdir, 'cert.pem'), 'w') as f:
            f.write("\n")

        rhsmcert.return_value.subject = {'CN': 123}
        rhsmconfig.return_value.get.side_effect = lambda group, key: {'consumerCertDir': cls.tempdir}.get(key, DEFAULT)
        cls.sm = SubscriptionManager(cls.logger, config)
        cls.sm.connection = MagicMock()
        cls.sm.connection.return_value.has_capability = MagicMock(return_value=False)
        cls.sm.connection.return_value.getConsumer = MagicMock(return_value={})
        cls.sm.connection.return_value.getOwner = MagicMock(return_value={'key': 'owner'})
        cls.uep_connection = patch('rhsm.connection.UEPConnection', cls.sm.connection)
        cls.uep_connection.start()
        cls.sm.cert_uuid = 123

    @classmethod
    def tearDownClass(cls):
        shutil.rmtree(cls.tempdir)
        cls.uep_connection.stop()

    def test_sendVirtGuests(self):
        config = VirtConfigSection.from_dict({'type': 'libvirt'}, 'test', None)
        report = DomainListReport(config, self.guestList, self.hypervisor_id)
        self.sm.sendVirtGuests(report)
        self.sm.connection.updateConsumer.assert_called_with(
            123,
            guest_uuids=[g.toDict() for g in self.guestList],
            hypervisor_id=self.hypervisor_id)

    def test_hypervisorCheckIn(self):
        owner = "owner"
        config = VirtConfigSection.from_dict({'type': 'libvirt', 'owner': owner}, 'test', None)
        # Ensure the data takes the proper for for the old API
        self.sm.connection.return_value.has_capability = MagicMock(return_value=False)
        self.sm.logger = MagicMock()
        report = HostGuestAssociationReport(config, self.mapping)
        self.sm.hypervisorCheckIn(report)
        self.sm.connection.hypervisorCheckIn.assert_called_with(
            owner,
            '',
            dict((host.hypervisorId, [g.toDict() for g in host.guestIds]) for host in self.mapping['hypervisors']),
            options=None)
        self.sm.logger.warning.assert_called_with("The hypervisor id '123' is assigned to 2 different systems. "
                        "Only one will be recorded at the server.")

    @patch('rhsm.connection.UEPConnection')
    # def test_hypervisorCheckInAsync(self):
    def test_hypervisorCheckInAsync(self, rhsmconnection):
        owner = 'owner'
        config = VirtConfigSection.from_dict({'type': 'libvirt', 'owner': owner}, 'test', None)
        # Ensure we try out the new API
        rhsmconnection.return_value.has_capability.return_value = True
        self.sm.logger = MagicMock()
        report = HostGuestAssociationReport(config, self.mapping)
        self.sm.hypervisorCheckIn(report)
        expected = {'hypervisors': [h.toDict() for h in self.mapping['hypervisors']]}
        self.sm.connection.hypervisorCheckIn.assert_called_with(
            'owner',
            '',
            expected,
            options=None
        )
        self.sm.logger.warning.assert_called_with("The hypervisor id '123' is assigned to 2 different systems. "
                        "Only one will be recorded at the server.")
        self.sm.connection.return_value.has_capability = MagicMock(return_value=False)

    @patch('rhsm.connection.UEPConnection')
    # def test_hypervisorHeartbeat(self):
    def test_hypervisorHeartbeat(self, rhsmconnection):
        owner = 'owner'
        config = VirtConfigSection.from_dict({'type': 'libvirt', 'owner': owner}, 'test', None)
        # Ensure we try out the new API
        rhsmconnection.return_value.has_capability.return_value = True
        self.sm.logger = MagicMock()
        self.sm.hypervisorHeartbeat(config, options=None)
        self.sm.connection.hypervisorHeartbeat.assert_called_with(
            owner,
            None
        )

    @patch('rhsm.connection.UEPConnection')
    def test_job_status(self, rhsmconnection):
        rhsmconnection.return_value.has_capability.return_value = True
        config = VirtConfigSection.from_dict({'type': 'libvirt', 'owner': 'owner'}, 'test', None)
        report = HostGuestAssociationReport(config, self.mapping)
        self.sm.hypervisorCheckIn(report)
        rhsmconnection.return_value.getJob.return_value = {
            'state': 'RUNNING',
        }
        self.sm.check_report_state(report)
        self.assertEqual(report.state, AbstractVirtReport.STATE_PROCESSING)

        def host(_host):
            return {
                'uuid': _host
            }

        rhsmconnection.return_value.getJob.return_value = {
            'state': 'FINISHED',
            'resultData': {
                'failedUpdate': ["failed"],
                'updated': [ {'uuid', '123'} ],
                'created': [ {'uuid' '456' } ],
                'unchanged': [ {'uuid', '789'} ]
            }
        }
        self.sm.logger = MagicMock()
        self.sm.check_report_state(report)
        # calls: authenticating + checking job status + 1 line about the number of unchanged
        self.assertEqual(self.sm.logger.debug.call_count, 3)
        self.assertEqual(report.state, AbstractVirtReport.STATE_FINISHED)

    @patch('rhsm.connection.UEPConnection')
    def test_job_status_with_status_command(self, rhsmconnection):
        rhsmconnection.return_value.has_capability.return_value = True
        config = VirtConfigSection.from_dict({'type': 'libvirt', 'owner': 'owner'}, 'test', None)
        report = StatusReport(config)
        rhsmconnection.return_value.getJob.return_value = {
            'state': 'RUNNING',
        }
        self.sm.check_report_state(report, status_call=True)
        self.assertEqual(report.state, AbstractVirtReport.STATE_PROCESSING)
        self.assertEqual(report.last_job_status, 'RUNNING')

        rhsmconnection.return_value.getJob.return_value = {
            'state': 'FINISHED',
            'resultData': {
                'failedUpdate': ["failed"],
                'updated': [ {'uuid', '123'} ],
                'created': [ {'uuid' '456' } ],
                'unchanged': [ {'uuid', '789'} ]
            }
        }
        self.sm.logger = MagicMock()
        self.sm.check_report_state(report, status_call=True)
        # calls: authenticating + checking job status + 1 line about the number of unchanged
        self.assertEqual(self.sm.logger.debug.call_count, 3)
        self.assertEqual(report.state, AbstractVirtReport.STATE_FINISHED)
        self.assertEqual(report.last_job_status, 'FINISHED')

class TestSubscriptionManagerConfig(TestBase):
    @classmethod
    @patch('rhsm.config.initConfig')
    @patch('rhsm.certificate.create_from_file')
    def setUpClass(cls, rhsmcert, rhsmconfig):
        super(TestSubscriptionManagerConfig, cls).setUpClass()
        options = Mock()
        cls.tempdir = tempfile.mkdtemp()
        with open(os.path.join(cls.tempdir, 'cert.pem'), 'w') as f:
            f.write("\n")
        rhsmcert.return_value.subject = {'CN': 123}
        rhsmconfig.return_value.get.side_effect = lambda group, key: {'consumerCertDir': cls.tempdir}.get(key, DEFAULT)
        cls.sm = SubscriptionManager(cls.logger, options)
        cls.sm.connection = MagicMock()
        cls.sm.connection.return_value.has_capability = MagicMock(return_value=False)
        cls.sm.connection.return_value.getConsumer = MagicMock(return_value={})
        cls.sm.connection.return_value.getOwner = MagicMock(return_value={'key': 'owner'})
        cls.uep_connection = patch('rhsm.connection.UEPConnection', cls.sm.connection)
        cls.uep_connection.start()
        cls.sm.cert_uuid = 123

    def setUp(self):
        self.config_dir = tempfile.mkdtemp()
        self.addCleanup(shutil.rmtree, self.config_dir)
        conf_dir_patch = patch('virtwho.config.VW_CONF_DIR', self.config_dir)
        conf_dir_patch.start()
        self.addCleanup(conf_dir_patch.stop)

    @classmethod
    def tearDownClass(cls):
        shutil.rmtree(cls.tempdir)
        cls.uep_connection.stop()

    def test_sm_config_env(self):
        os.environ = {
            "VIRTWHO_SAM": '1',
            "VIRTWHO_LIBVIRT": '1'
        }
        sys.argv = ["virt-who"]
        logger, config = parse_options()
        manager = Manager.from_config(logger, config)
        self.assertTrue(isinstance(manager, SubscriptionManager))

    @pytest.mark.skipif(not six.PY2, reason="test only runs with python 2 virt-who")
    def test_sm_config_cmd(self):
        os.environ = {}
        sys.argv = ["virt-who", "--sam", "--libvirt"]
        logger, effective_config = parse_options()
        config_manager = DestinationToSourceMapper(effective_config)
        self.assertEqual(len(config_manager.configs), 1)
        config = dict(config_manager.configs)[VW_ENV_CLI_SECTION_NAME]
        manager = Manager.from_config(self.logger, config)
        self.assertTrue(isinstance(manager, SubscriptionManager))

    def test_sm_config_file(self):
        config_dir = tempfile.mkdtemp()
        self.addCleanup(shutil.rmtree, config_dir)
        with open(os.path.join(config_dir, "test.conf"), "w") as f:
            f.write("""
[test]
type=libvirt
owner=owner
rhsm_hostname=host
rhsm_port=8080
rhsm_prefix=prefix
rhsm_proxy_hostname=proxy_host
rhsm_proxy_port=9090
rhsm_proxy_user=proxy_user
rhsm_proxy_password=proxy_password
rhsm_insecure=1
rhsm_username=user
rhsm_password=passwd
rhsm_no_proxy=filter
""")

        config_manager = DestinationToSourceMapper(init_config({}, config_dir=config_dir))
        self.assertEqual(len(config_manager.configs), 1)
        config = dict(config_manager.configs)["test"]
        manager = Manager.from_config(self.logger, config)
        self.assertTrue(isinstance(manager, SubscriptionManager))
        self.assertEqual(config['rhsm_hostname'], 'host')
        self.assertEqual(config['rhsm_port'], '8080')

        manager._connect(config)
        self.sm.connection.assert_called_with(
            username='user',
            password='passwd',
            host='host',
            ssl_port=8080,
            handler='prefix',
            proxy_hostname='proxy_host',
            proxy_port='9090',
            proxy_user='proxy_user',
            proxy_password='proxy_password',
            no_proxy='filter',
            insecure='1',
            correlation_id=manager.correlation_id)

    @unittest.skip("skip until rhsm is fixed")
    @patch('rhsm.connection.RhsmProxyHTTPSConnection')
    @patch('M2Crypto.httpslib.HTTPSConnection')
    @patch('rhsm.config.initConfig')
    def test_sm_config_override(self, initConfig, HTTPSConnection, RhsmProxyHTTPSConnection):
        """Test if overriding options from rhsm.conf works."""

        conn = MagicMock()
        conn.getresponse.return_value.status = 200
        conn.getresponse.return_value.read.return_value = '{"result": "ok"}'
        HTTPSConnection.return_value = conn
        RhsmProxyHTTPSConnection.return_value = conn

        def config_get(section, key):
            return {
                'server/proxy_hostname': 'proxy.server.test',
                'rhsm/consumerCertDir': '',
                'server/hostname': 'server.test',
                'server/port': '8081',
                'server/prefix': 'old_prefix',
            }.get('%s/%s' % (section, key), None)
        initConfig.return_value.get.side_effect = config_get
        config_dir = tempfile.mkdtemp()
        self.addCleanup(shutil.rmtree, config_dir)
        with open(os.path.join(config_dir, "test.conf"), "w") as f:
            f.write("""
[test]
type=libvirt
owner=owner
rhsm_hostname=host
rhsm_port=8080
rhsm_prefix=/prefix
rhsm_proxy_hostname=
rhsm_proxy_port=8443
rhsm_insecure=1
rhsm_username=user
rhsm_password=passwd
""")

        conf = parse_file(os.path.join(config_dir, "test.conf"))
        effective_config = EffectiveConfig()
        conf_values = conf.pop("test")
        effective_config["test"] = VirtConfigSection.from_dict(
            conf_values,
            "test",
            effective_config
        )
        config_manager = DestinationToSourceMapper(effective_config)
        self.assertEqual(len(config_manager.configs), 1)
        config = config_manager.configs[0]
        manager = Manager.fromOptions(self.logger, Mock(), config)
        self.assertTrue(isinstance(manager, SubscriptionManager))
        self.assertEqual(config.rhsm_hostname, 'host')
        self.assertEqual(config.rhsm_port, '8080')

        manager._connect(config)
        self.assertFalse(RhsmProxyHTTPSConnection.called, "It shouldn't use proxy")
        self.assertTrue(HTTPSConnection.called)
        conn.request.assert_called_with(
            'GET',
            '/prefix/status/',
            body=ANY,
            headers=ANY)



from __future__ import print_function

#
# Module for abstraction of all virtualization backends, part of virt-who
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#

import os
from multiprocessing import Process
from six.moves.SimpleHTTPServer import SimpleHTTPRequestHandler
import random
from six.moves import socketserver
import requests
import json


class FakeTCPServer(socketserver.TCPServer):
    allow_reuse_address = True


class FakeHandler(SimpleHTTPRequestHandler):
    """
    Base class of handler for:
    - Fake virt backends like RHEVM or ESX
    - Fake candlepin server (FakeSam)
    """
    def write_file(self, directory, filename):
        """
        Send file with given `filename` to the client. File must be in
        `directory` in the data/ subdirectory of the directory
        where the current __file__ is.
        """
        base = os.path.dirname(os.path.abspath(__file__))
        path = os.path.join(base, 'data', directory, filename)

        self.send_response(200)
        self.send_header("Content-type", "text/html")
        self.send_header("Content-length", os.stat(path).st_size)
        self.end_headers()
        with open(path, "rb") as f:
            self.wfile.write(f.read())

    def write_json(self, content, status_code=requests.codes.ok, headers=None):
        """
        Send back as json the content parameter (Most useful to input a dict).
        Sets the following headers if not provided (other wise they are added to the input headers)
          "Content-type" : "application/json"
          "Content-length" : len(json.dumps(content).encode('utf-8'))
        :param content: The object to convert to json and send as the body of the message
        :type content: dict

        :param status_code: The HTTP Status code to set on the response
        :type status_code: int

        :param headers: A dict of header_name : header_value.
        :type headers: dict
        :return: Nothing
        :rtype: None
        """

        if headers is None:
            headers = {}
        if "Content-type" not in headers:
            headers["Content-type"] = "application/json"

        content = json.dumps(content).encode('utf-8')
        if "Content-length" not in headers:
            headers["Content-length"] = len(content)

        self.send_response(status_code)
        for header, value in headers.items():
            self.send_header(header, value)
        self.end_headers()
        self.wfile.write(content)


class FakeServer(Process):
    """
    Base class for fake servers like:
    - Fake virt backends like RHEVM or ESX
    - Fake candlepin server
    """

    def __init__(self, handler_class, host='localhost', port=None):
        super(FakeServer, self).__init__()
        self._port = port
        self.host = host
        self.server = FakeTCPServer((self.host, self.port), handler_class)

    @property
    def port(self):
        if self._port is None:
            self._port = random.randint(8000, 9000)
        return self._port

    def clear_port(self):
        print("Clear port: ", self._port)
        self._port = None

    def run(self):
        for i in range(100):
            try:
                print("Starting {cls} on {host}:{port}".format(
                    cls=self.__class__.__name__,
                    host=self.host,
                    port=self.port)
                )
                self.server.serve_forever()
                break
            except AssertionError:
                self.clear_port()
        else:
            raise AssertionError("No free port found, starting aborted")



from __future__ import print_function
from base import TestBase

from mock import patch, Mock, sentinel, call
from virtwho.virt import IntervalThread
from threading import Event
from datetime import datetime

class TestIntervalThreadTiming(TestBase):
    """
    This is a group of tests intended to test the timing of the interval thread.
    """

    def setUp(self):
        time_patcher = patch('time.sleep')
        self.mock_time = time_patcher.start()
        event_patcher = patch('virtwho.virt.virt.Event')
        self.mock_internal_terminate_event = event_patcher.start().return_value
        self.mock_internal_terminate_event.is_set.return_value = False
        self.addCleanup(time_patcher.stop)
        self.addCleanup(event_patcher.stop)

        # default mock objects that can be passed into a new interval thread
        self.config = Mock()
        self.source = Mock()
        self.dest = Mock()
        self.terminate_event = Mock()
        self.terminate_event.is_set.return_value = False
        self.interval = 2
        self.oneshot = True

        self._get_data_return = sentinel.default_data_to_send
        self._send_data_return = sentinel.default_send_data_return

    def setup_interval_thread(self, **kwargs):
        """
        Sets up an interval thread class with mocks of unimplemented methods
        to allow testing of the base implementation of those methods that have
        concrete implementations.
        """
        logger = kwargs.get('logger', self.logger)
        config = kwargs.get('config', self.config)
        source = kwargs.get('source', self.source)
        dest = kwargs.get('dest', self.dest)
        terminate_event = kwargs.get('terminate_event', self.terminate_event)
        interval = kwargs.get('interval', self.interval)
        oneshot = kwargs.get('oneshot', self.oneshot)

        interval_thread = IntervalThread(logger, config, source, dest,
                                         terminate_event, interval, oneshot)

        mock_get_data_impl = kwargs.get('mock_get_data', None)
        if mock_get_data_impl is None:
            mock_get_data_impl = Mock()
            mock_get_data_impl.return_value = self._get_data_return

        mock_send_data_impl = kwargs.get('mock_send_data', None)
        if mock_send_data_impl is None:
            mock_send_data_impl = Mock()
            mock_send_data_impl.return_value = self._send_data_return

        interval_thread._get_data = mock_get_data_impl
        interval_thread._send_data = mock_send_data_impl

        return interval_thread


    def test__run(self):
        """
        Tests the timing of the _run method
        """
        oneshot = False
        interval = 20  # Seconds
        start_time = datetime(1, 1, 1, 1, 1)
        time_taken = 10  # seconds, must be less than 60
        end_time = datetime(start_time.year, start_time.month,
                            start_time.day, start_time.hour, start_time.minute,
                            start_time.second + time_taken,
                            start_time.microsecond, start_time.tzinfo)
        expected_wait_time = interval - time_taken
        with patch('virtwho.virt.virt.datetime') as patcher:
            patcher.now = Mock(side_effect=[start_time, end_time])
            interval_thread = self.setup_interval_thread(oneshot=oneshot,
                                                         interval=interval)
            interval_thread.is_terminated = Mock(side_effect=[False, True])
            interval_thread.wait = Mock()
            interval_thread._run()
            interval_thread._get_data.assert_called()
            interval_thread._send_data.assert_has_calls([call(data_to_send=self._get_data_return)])
            interval_thread.wait.assert_has_calls([call(expected_wait_time)])

    def test__run_send_takes_longer_than_interval(self):
        """
        Tests the timing of the _run method
        """
        oneshot = False
        interval = 20  # Seconds
        start_time = datetime(1, 1, 1, 1, 1)
        time_taken = 21  # seconds, must be less than 60
        end_time = datetime(start_time.year, start_time.month,
                            start_time.day, start_time.hour, start_time.minute,
                            start_time.second + time_taken,
                            start_time.microsecond, start_time.tzinfo)
        with patch('virtwho.virt.virt.datetime') as patcher:
            patcher.now = Mock(side_effect=[start_time, end_time])
            interval_thread = self.setup_interval_thread(oneshot=oneshot,
                                                         interval=interval)
            interval_thread.is_terminated = Mock(side_effect=[False, True])
            interval_thread.wait = Mock()
            interval_thread._run()
            interval_thread._get_data.assert_called()
            interval_thread._send_data.assert_has_calls([call(data_to_send=self._get_data_return)])
            interval_thread.wait.assert_not_called()


    def test_wait(self):
        interval_thread = self.setup_interval_thread()
        interval_thread.is_terminated = Mock()
        interval_thread.is_terminated.return_value = False
        # The total time we expect to be waited
        wait_time = 10
        # The time we expect to be waited each interval
        expected_wait_interval = 1
        expected_calls = [call(expected_wait_interval) for x in range(wait_time)]

        interval_thread.wait(wait_time=wait_time)
        self.assertEqual(interval_thread.is_terminated.call_count, wait_time)
        self.mock_time.assert_has_calls(expected_calls)

    def test_is_terminated_terminate_event(self):
        interval_thread = self.setup_interval_thread()
        self.assertEqual(False, interval_thread.is_terminated())
        self.terminate_event.is_set.return_value = True
        self.assertEqual(True, interval_thread.is_terminated())

    def test_is_terminated_internal_terminate_event(self):
        interval_thread = self.setup_interval_thread()
        self.assertEqual(False, interval_thread.is_terminated())
        self.mock_internal_terminate_event.is_set.return_value = True
        self.assertEqual(True, interval_thread.is_terminated())

    def test_is_terminated_both_events(self):
        interval_thread = self.setup_interval_thread()
        self.assertEqual(False, interval_thread.is_terminated())
        interval_thread._internal_terminate_event.is_set.return_value = True
        self.terminate_event.is_set.return_value = True
        self.assertEqual(True, interval_thread.is_terminated())

    def test_stop(self):
        interval_thread = self.setup_interval_thread()
        interval_thread.stop()
        self.mock_internal_terminate_event.set.assert_called()

    def test_run(self):
        oneshot = False
        interval = 20  # Seconds
        interval_thread = self.setup_interval_thread(oneshot=oneshot,
                                                     interval=interval)
        interval_thread.is_terminated = Mock(side_effect=[False, False, True])
        interval_thread._run = Mock()
        interval_thread.wait = Mock()
        interval_thread.run()
        interval_thread.wait.assert_has_calls([call(interval)])

    def test_run_has_error(self):
        oneshot = False
        interval = 20  # Seconds
        interval_thread = self.setup_interval_thread(oneshot=oneshot,
                                                     interval=interval)
        interval_thread.is_terminated = Mock(side_effect=[False, False,
                                                          False, True])
        interval_thread._run = Mock(side_effect=Exception)
        interval_thread.wait = Mock()
        interval_thread.run()
        interval_thread.wait.assert_has_calls([call(interval)])

    def test_run_has_error_and_terminated(self):
        oneshot = False
        interval = 20  # Seconds
        interval_thread = self.setup_interval_thread(oneshot=oneshot,
                                                     interval=interval)
        interval_thread.is_terminated = Mock(side_effect=[False, True,
                                                          True, True])
        interval_thread._run = Mock(side_effect=Exception)
        interval_thread.wait = Mock()
        interval_thread.run()
        interval_thread.wait.assert_not_called()



# coding=utf-8

"""
Test reading and writing configuration files as well as configuration objects.

Copyright (C) 2014 Radek Novacek <rnovacek@redhat.com>

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""

import os
import shutil
import tempfile

from virtwho.migrate.migrateconfiguration import migrate_env_to_config

from base import TestBase, unittest


class TestConfigMigrate(TestBase):

    tmp_config_dir = tempfile.TemporaryDirectory().name
    virt_who_filename = os.path.join(tmp_config_dir, "virt-who")
    virt_who_conf_filename = os.path.join(tmp_config_dir, "virt-who.conf")

    def setUp(self):
        if not os.path.exists(self.tmp_config_dir):
            os.makedirs(self.tmp_config_dir)

    def tearDown(self):
        if os.path.exists(self.tmp_config_dir):
            shutil.rmtree(self.tmp_config_dir)

    def test_migrate_not_existing(self):
        sysconfig =\
"""VIRTWHO_INTERVAL=60
VIRTWHO_DEBUG=1
VIRTWHO_ONE_SHOT=0
HTTPS_PROXY=one
HTTP_PROXY=two
NO_PROXY=*
"""
        expected_conf =\
"""[global]
#migrated
interval=60
#migrated
debug=True
#migrated
oneshot=False

[system_environment]
#migrated
HTTPS_PROXY=one
#migrated
HTTP_PROXY=two
#migrated
NO_PROXY=*
"""

        with open(self.virt_who_filename, "w") as f:
            f.writelines(sysconfig)
        migrate_env_to_config(self.virt_who_filename,
                              self.virt_who_conf_filename)
        with open(self.virt_who_conf_filename, "r") as conf:
            result = conf.readlines()
        self.assertEqual("".join(result), expected_conf)


    def test_migrate_to_existing_no_env(self):
            sysconfig = \
"""VIRTWHO_INTERVAL=60
VIRTWHO_DEBUG=1
VIRTWHO_ONE_SHOT=0
"""
            existing = \
"""[global]
reporter_id=this_one

"""
            expected = \
"""[global]
#migrated
interval=60
#migrated
debug=True
#migrated
oneshot=False
reporter_id=this_one

"""

            with open(self.virt_who_filename, "w") as f:
                f.writelines(sysconfig)
            with open(self.virt_who_conf_filename, "w") as f:
                f.writelines(existing)
            migrate_env_to_config(self.virt_who_filename,
                                  self.virt_who_conf_filename)
            with open(self.virt_who_conf_filename, "r") as conf:
                result = conf.readlines()
            self.assertEqual("".join(result), expected)


    def test_migrate_to_existing(self):
        sysconfig = \
"""VIRTWHO_INTERVAL=60
VIRTWHO_DEBUG=1
VIRTWHO_ONE_SHOT=0
HTTPS_PROXY=one
HTTP_PROXY=two
"""
        existing =\
"""[global]
reporter_id=this_one

[system_environment]
no_proxy=*
"""
        expected =\
"""[global]
#migrated
interval=60
#migrated
debug=True
#migrated
oneshot=False
reporter_id=this_one

[system_environment]
#migrated
HTTPS_PROXY=one
#migrated
HTTP_PROXY=two
no_proxy=*
"""

        with open(self.virt_who_filename, "w") as f:
            f.writelines(sysconfig)
        with open(self.virt_who_conf_filename, "w") as f:
            f.writelines(existing)
        migrate_env_to_config(self.virt_who_filename,
                              self.virt_who_conf_filename)
        with open(self.virt_who_conf_filename, "r") as conf:
            result = conf.readlines()
        self.assertEqual("".join(result), expected)


    def test_migrate_to_existing_entries(self):
        sysconfig = \
"""VIRTWHO_INTERVAL=60
"""
        existing =\
"""[global]
interval=360
reporter_id=this_one

[system_environment]
no_proxy=*
"""
        expected =\
"""[global]
#migrated
interval=60
interval=360
reporter_id=this_one

[system_environment]
no_proxy=*
"""

        with open(self.virt_who_filename, "w") as f:
            f.writelines(sysconfig)
        with open(self.virt_who_conf_filename, "w") as f:
            f.writelines(existing)
        migrate_env_to_config(self.virt_who_filename,
                              self.virt_who_conf_filename)
        with open(self.virt_who_conf_filename, "r") as conf:
            result = conf.readlines()
        self.assertEqual("".join(result), expected)


    def test_migrate_to_no_sysconfig(self):
        existing =\
"""[global]
interval=360
reporter_id=this_one

[system_environment]
no_proxy=*
"""
        expected =\
"""[global]
interval=360
reporter_id=this_one

[system_environment]
no_proxy=*
"""

        with open(self.virt_who_conf_filename, "w") as f:
            f.writelines(existing)
        migrate_env_to_config(self.virt_who_filename,
                              self.virt_who_conf_filename)
        with open(self.virt_who_conf_filename, "r") as conf:
            result = conf.readlines()
        self.assertEqual("".join(result), expected)


    def test_migrate_to_existing_commented_header(self):
        sysconfig = \
"""VIRTWHO_INTERVAL=60
VIRTWHO_DEBUG=1
VIRTWHO_ONE_SHOT=0
HTTPS_PROXY=one
HTTP_PROXY=two
"""
        existing =\
"""#[system_environment]   ; Values in this section will be copied to the system environment and will be in place
#                       ; for the extent of the execution.
#                       ; There are a few suggested options, but any name/value intended for the system environment
#                       ; can be put here. Other than the copy process, virt-who will not read or specifically react
#                       ; to them.

# [global]
# reporter_id=this_one

#[system_environment]
"""
        expected =\
"""#[system_environment]   ; Values in this section will be copied to the system environment and will be in place
#                       ; for the extent of the execution.
#                       ; There are a few suggested options, but any name/value intended for the system environment
#                       ; can be put here. Other than the copy process, virt-who will not read or specifically react
#                       ; to them.

[global]
#migrated
interval=60
#migrated
debug=True
#migrated
oneshot=False
# reporter_id=this_one

[system_environment]
#migrated
HTTPS_PROXY=one
#migrated
HTTP_PROXY=two
"""
        self.maxDiff = None
        with open(self.virt_who_filename, "w") as f:
            f.writelines(sysconfig)
        with open(self.virt_who_conf_filename, "w") as f:
            f.writelines(existing)
        migrate_env_to_config(self.virt_who_filename,
                              self.virt_who_conf_filename)
        with open(self.virt_who_conf_filename, "r") as conf:
            result = conf.readlines()
        self.assertEqual("".join(result), expected)



# -*- coding: utf-8 -*-
from __future__ import print_function, absolute_import
"""
Agent for reporting virtual guest IDs to subscription-manager

Copyright (C) 2011 Radek Novacek <rnovacek@redhat.com>

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""

import sys
import os
import time
import fcntl

PIDFILE = "/var/run/virt-who.pid"
STATUS_LOCK = "/var/run/virt-who-status.pid"
STATUS_DATA = "/var/lib/virt-who/run_data.json"
STATUS_DATA_DIR = "/var/lib/virt-who"
DEFAULT_TIMEOUT = 1.0


class PIDLock(object):
    def __init__(self, filename):
        self.filename = filename

    def is_locked(self):
        try:
            with open(self.filename, "r") as f:
                pid = int(f.read().strip())
            try:
                # this does not kill the process, it is a way
                # to see if the process with this PID still exists
                os.kill(pid, 0)
                return True
            except OSError:
                # Process no longer exists
                print("PID file exists but associated process " \
                                     "does not, deleting PID file", file=sys.stderr)
                os.remove(self.filename)
                return False
        except Exception:
            return False

    def __enter__(self):
        # Write pid to pidfile
        try:
            with os.fdopen(
                    os.open(self.filename, os.O_WRONLY | os.O_CREAT, 0o600),
                    'w') as f:
                f.write("%d" % os.getpid())
        except Exception as e:
            print("Unable to create pid file: %s" % str(e), file=sys.stderr)

    def __exit__(self, exc_type, exc_value, traceback):
        try:
            os.remove(self.filename)
        except Exception:
            pass

class FileLock(object):
    def __init__(self, filename, logger=None, timeout=None):
        self.lock_file = filename
        self.logger = logger
        self.timeout = timeout or DEFAULT_TIMEOUT

    def __enter__(self):
        open_mode = os.O_RDWR | os.O_CREAT | os.O_TRUNC
        fd = os.open(self.lock_file, open_mode)
        self.lock_file_fd = None

        start_time = current_time = time.time()
        while current_time < start_time + self.timeout:
            try:
                # The LOCK_EX means that only one process can hold the lock
                # The LOCK_NB means that the fcntl.flock() is not blocking
                # and we are able to implement termination of while loop,
                # when timeout is reached.
                # More information here:
                # https://docs.python.org/3/library/fcntl.html#fcntl.flock
                fcntl.flock(fd, fcntl.LOCK_EX | fcntl.LOCK_NB)
            except (IOError, OSError):
                pass
            else:
                self.lock_file_fd = fd
                break
            time.sleep(0.2)
            current_time = time.time()
        # we did not get the lock in the time allowed
        if self.lock_file_fd is None:
            os.close(fd)
            raise IOError

    def __exit__(self, exc_type, exc_value, traceback):
        # Do not remove the lockfile:
        #
        #   https://github.com/benediktschmitt/py-filelock/issues/31
        #   https://stackoverflow.com/questions/17708885/flock-removing-locked-file-without-race-condition
        fcntl.flock(self.lock_file_fd, fcntl.LOCK_UN)
        os.close(self.lock_file_fd)



#!/usr/bin/python2
# -*- coding: utf-8 -*-
from __future__ import print_function

import os
import sys
from bs4 import BeautifulSoup


class InvalidXmlError(Exception):
    pass


already_included = set()


def process_file(filename):
    xml = BeautifulSoup(open(filename), "xml")
    for include in xml.find_all(['include', 'import']):
        location = None
        for attr, value in include.attrs.items():
            if attr.lower() == 'schemalocation':
                location = value
            if attr.lower() == 'location':
                location = value
        if not location:
            raise InvalidXmlError("No schemalocation attribute: %s" % (include.prettify()))
        if location in already_included:
            include.decompose()
            continue
        else:
            already_included.add(location)
        filename = os.path.join(os.path.dirname(vimfile), location)
        inc_xml = process_file(filename)
        for tag in inc_xml.find_all(True, recursive=False):
            if include.parent.name == tag.name:
                for subtag in tag.find_all(True, recursive=False):
                    include.insert_before(subtag)
            else:
                include.insert_before(tag)
        include.decompose()
    return xml


def clean_up_vim(vim, keep_methods=None, keep_types=None):
    if keep_types is None:
        keep_types = set()
    keep_messages = set()
    for operation in vim.binding.find_all('operation', recursive=False):
        if operation['name'] in keep_methods:
            pass
        else:
            operation.decompose()

    for operation in vim.portType.find_all('operation', recursive=False):
        if operation['name'] in keep_methods:
            for io in operation.find_all(['input', 'output'], recursive=False):
                keep_messages.add(io['message'].rpartition(":")[2])
            for fault in operation.find_all('fault', recursive=False):
                keep_messages.add(fault['message'].rpartition(":")[2])
                keep_types.add(fault['name'])
        else:
            operation.decompose()

    print("Keep messages:", keep_messages, file=sys.stderr)
    for message in vim.definitions.find_all('message', recursive=False):
        if message['name'] in keep_messages:
            for part in message.find_all('part', recursive=False):
                keep_types.add(part['element'].rpartition(":")[2])
        else:
            message.decompose()

    # Gather recursive types
    changed = True
    while changed:
        changed = False
        for type in vim.types.schema.find_all(True, recursive=False):
            if 'name' in type.attrs and type['name'] in keep_types:
                if 'type' in type.attrs:
                    t = type.attrs['type'].rpartition(":")[2]
                    if t not in keep_types:
                        changed = True
                    keep_types.add(t)
                for ext in type.find_all('extension'):
                    t = ext['base'].rpartition(":")[2]
                    if t not in keep_types:
                        changed = True
                    keep_types.add(t)
                for element in type.find_all('element'):
                    t = element['type'].rpartition(":")[2]
                    if t not in keep_types:
                        changed = True
                    keep_types.add(t)

    print("Keep types:", keep_types, file=sys.stderr)
    for type in vim.types.schema.find_all(True, recursive=False):
        if 'name' not in type.attrs or type['name'] not in keep_types:
            type.decompose()

    return vim

if __name__ == '__main__':
    if len(sys.argv) < 2:
        print("Usage: %s /path/to/vim.wsdl")
        sys.exit(1)

    vimfile = sys.argv[1]

    # replace the includes
    vim = process_file(vimfile)
    vim.definitions.attrs['xmlns:vim25'] = "urn:vim25"
    vim.definitions.attrs['targetNamespace'] = "urn:vim25"
    vim.definitions.attrs['xmlns:mime'] = "http://schemas.xmlsoap.org/wsdl/mime/"
    vim.definitions.attrs['xmlns:xsd'] = "http://www.w3.org/2001/XMLSchema"
    with open('/tmp/vim_full.xml', 'w') as f:
        f.write(vim.prettify())
    filtered_vim = clean_up_vim(
        vim,
        keep_methods=set((
            'Login', 'RetrieveServiceContent', 'RetrieveProperties',
            'RetrievePropertiesEx', 'CreateFilter', 'WaitForUpdatesEx',
            'DestroyPropertyFilter', 'CancelWaitForUpdates')),
        keep_types=set((
            'TraversalSpec', 'ArrayOfManagedObjectReference',
            'ArrayOfDynamicProperty', 'DynamicData', 'VimFault',
            'PropertyFilterSpec')))
    '''ManagedObjectReference', 'ServiceContent',
    'AboutInfo', 'UserSession', 'ObjectSpec',
    'DynamicProperty', 'TraversalSpec',
    'PropertyFilterSpec', 'PropertySpec'''
    print(filtered_vim.prettify())



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function
import time
import json
import os
from json.decoder import JSONDecodeError
from threading import Event

from virtwho import log

from virtwho.config import DestinationToSourceMapper, VW_GLOBAL
from virtwho.datastore import Datastore
from virtwho.manager import Manager
from virtwho.lock import FileLock, STATUS_LOCK, STATUS_DATA, STATUS_DATA_DIR
from virtwho.virt import Virt, info_to_destination_class

try:
    from collections import OrderedDict
except ImportError:
    # Python 2.6 doesn't have OrderedDict, we need to have our own
    from .util import OrderedDict


class ReloadRequest(Exception):
    ''' Reload of virt-who was requested by sending SIGHUP signal. '''


class ExitRequest(Exception):
    """ Indicates that something that should cause virt-who to exit has occurred"""

    def __init__(self, code=0, message=""):
        self.code = code
        self.message = message


class Executor(object):
    def __init__(self, logger, options):
        """
        Executor class provides bridge between virtualization supervisor and
        Subscription Manager.

        logger - logger instance
        options - options for virt-who, parsed from command line arguments
        """
        self.logger = logger
        self.options = options
        self.terminate_event = Event()
        self.virts = []
        self.destinations = []

        # Queue for getting events from virt backends
        self.datastore = Datastore()
        self.reloading = False

        self.dest_to_source_mapper = DestinationToSourceMapper(options)

        for name, config in self.dest_to_source_mapper.configs:
            logger.info("Using config named '%s'" % name)

    def _create_virt_backends(self):
        """
        Create virts list with virt backend threads
        """
        virts = []
        config_names = []
        for name, config in self.dest_to_source_mapper.configs:
            config_names.append(name)
            try:
                virt = Virt.from_config(self.logger, config, self.datastore,
                                        terminate_event=self.terminate_event,
                                        interval=self.options[VW_GLOBAL]['interval'],
                                        oneshot=self.options[VW_GLOBAL]['oneshot'],
                                        status=self.options[VW_GLOBAL]['status'])
            except Exception as e:
                self.logger.error('Unable to use configuration "%s": %s', name, str(e))
                continue
            virts.append(virt)
        self._init_run_status(config_names)
        return virts

    def _init_run_status(self, config_names=[]):
        # ensure that there is an entry in the status file for each config. We don't care if they work, we
        # need to record their existence
        need_write = False
        try:
            with FileLock(STATUS_LOCK):
                os.makedirs(STATUS_DATA_DIR, exist_ok=True)
                if os.path.exists(STATUS_DATA):
                    with open(STATUS_DATA, "r") as json_status:
                        try:
                            status_dict = json.load(json_status)
                        except JSONDecodeError as e:
                            status_dict = {}
                else:
                    status_dict = {}
                if 'sources' not in status_dict:
                    status_dict['sources'] = {}
                    need_write = True
                if 'destinations' not in status_dict:
                    status_dict['destinations'] = {}
                    need_write = True
                for name in config_names:
                    if name not in status_dict['sources']:
                        status_dict['sources'][name] = {"last_successful_retrieve": None}
                        need_write = True
                    if name not in status_dict['destinations']:
                        status_dict['destinations'][name] = {"last_successful_send": None,"last_job_id": None}
                        need_write = True
                if need_write:
                    # need to create the file if it does not exist
                    with open(STATUS_DATA, "w+") as json_status:
                        json.dump(status_dict, json_status)
        except IOError:
            self.logger.error("Unable to record run data. Cannot get lock on file.")

    def _create_destinations(self):
        """Populate self.destinations with a list of  list with them

            @param reset: Whether to kill existing destinations or not, defaults
            to false
            @type: bool
        """
        dests = []
        for info in self.dest_to_source_mapper.dests:
            # Dests should already include all destinations we want created
            # at this time. This method will make no assumptions of creating
            # defaults of any kind.
            source_keys = self.dest_to_source_mapper.dest_to_sources_map[info]
            info.name = "destination_%s" % hash(info)
            logger = log.getLogger(name=info.name)
            manager = Manager.fromInfo(logger, self.options, info)
            dest_class = info_to_destination_class[type(info)]
            dest = dest_class(config=info, logger=logger,
                              source_keys=source_keys,
                              options=self.options,
                              source=self.datastore, dest=manager,
                              terminate_event=self.terminate_event,
                              interval=self.options[VW_GLOBAL]['interval'],
                              oneshot=self.options[VW_GLOBAL]['oneshot'],
                              status=self.options[VW_GLOBAL]['status'])
            dests.append(dest)
        return dests

    @staticmethod
    def wait_on_threads(threads, max_wait_time=None, kill_on_timeout=False):
        """
        Wait for each of the threads in the list to be terminated
        @param threads: A list of IntervalThread objects to wait on
        @type threads: list

        @param max_wait_time: An optional max amount of seconds to wait
        @type max_wait_time: int

        @param kill_on_timeout: An optional arg that, if truthy and
        max_wait_time is defined and exceeded, cause this method to attempt
        to terminate and join the threads given it.
        @type kill_on_timeout: bool

        @return: A list of threads that have not quit yet. Without a
        max_wait_time this list is always empty (or we are stuck waiting).
        With a max_wait_time this list will include those threads that have
        not quit yet.
        @rtype: list
        """
        delta_time = 1.0
        total_waited = 0
        threads_not_terminated = list(threads)
        while len(threads_not_terminated) > 0:
            if max_wait_time is not None and total_waited > max_wait_time:
                if kill_on_timeout:
                    Executor.terminate_threads(threads_not_terminated)
                    return []
                return threads_not_terminated
            for thread in threads_not_terminated:
                if thread.is_terminated():
                    threads_not_terminated.remove(thread)
            if not threads_not_terminated:
                break
            time.sleep(delta_time)
            if max_wait_time is not None:
                total_waited += 1 * 1.0/delta_time
        return threads_not_terminated

    @staticmethod
    def terminate_threads(threads):
        for thread in threads:
            thread.stop()
            if thread.ident:
                thread.join()

    def run_oneshot(self):
        # Start all sources
        self.virts = self._create_virt_backends()

        if len(self.virts) == 0:
            err = "virt-who can't be started: no suitable virt backend found"
            self.logger.error(err)
            raise ExitRequest(code=1, message=err)

        self.destinations = self._create_destinations()

        if len(self.destinations) == 0:
            err = "virt-who can't be started: no suitable destinations found"
            self.logger.error(err)
            raise ExitRequest(code=1, message=err)

        for thread in self.virts:
            thread.start()

        Executor.wait_on_threads(self.virts)

        if self.options[VW_GLOBAL]['print']:
            to_print = {}
            for source in self.dest_to_source_mapper.sources:
                try:
                    report = self.datastore.get(source)
                    config = report.config
                    to_print[config.name] = report
                except KeyError:
                    self.logger.info(f"Unable to retrieve report for source '{source}' for printing")
            return to_print

        for thread in self.destinations:
            thread.start()

        Executor.wait_on_threads(self.destinations)
        if self.options[VW_GLOBAL]['status']:
            output = {}
            for source in self.dest_to_source_mapper.sources:
                try:
                    report = self.datastore.get(source)
                except KeyError:
                    self.logger.info(f"Unable to retrieve report for source '{source}' for printing")
                else:
                    output[report.config.name] = report
            return output

    def run(self):
        self.logger.debug("Starting infinite loop with %d seconds interval",
                          self.options[VW_GLOBAL]['interval'])

        # Need to update the dest to source mapping of the dest_to_source_mapper object
        # here because of the way that main reads the config from the command
        # line
        # TODO: Update dests to source map on addition or removal of configs
        self.dest_to_source_mapper.update_dest_to_source_map()
        # Start all sources
        self.virts = self._create_virt_backends()

        if len(self.virts) == 0:
            err = "virt-who can't be started: no suitable virt backend found"
            self.logger.error(err)
            raise ExitRequest(code=1, message=err)

        self.destinations = self._create_destinations()
        if len(self.destinations) == 0:
            err = "virt-who can't be started: no suitable destinations found"
            self.logger.error(err)
            raise ExitRequest(code=1, message=err)

        for thread in self.virts:
            thread.start()

        for thread in self.destinations:
            thread.start()

        # Interruptibly wait on the other threads to be terminated
        self.wait_on_threads(self.destinations)

        raise ExitRequest(code=0)

    def stop_threads(self):
        self.terminate_event.set()
        self.terminate_threads(self.virts)
        self.terminate_threads(self.destinations)

    def terminate(self):
        self.logger.debug("virt-who is shutting down")
        self.stop_threads()
        self.virts = []
        self.destinations = []
        self.datastore = None

    def reload(self):
        """
        Causes all threads to be terminated in preparation for running again
        """
        self.stop_threads()
        self.terminate_event.clear()
        self.datastore = Datastore()



# -*- coding: utf-8 -*-
from __future__ import print_function, absolute_import

#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#

import json

from virtwho.config import VirtConfigSection, accessible_file, str_to_bool
from virtwho.virt import Virt, VirtError, Guest, Hypervisor
from virtwho.util import decode


class FakeVirtConfigSection(VirtConfigSection):
    """
    This class is used for validation of fake virtualization backend
    section(s). It tries to validate options and combination of options that
    are specific for this virtualization backend. In specific, it attempts to read
    the given file and produces error messages if it is not usable.
    """
    VIRT_TYPE = 'fake'

    def __init__(self, *args, **kwargs):
        super(FakeVirtConfigSection, self).__init__(*args, **kwargs)
        self.add_key('is_hypervisor', validation_method=self._validate_str_to_bool, default=True,
                     required=True)
        self.add_key('file', validation_method=self._validate_fake_virt_file, required=True)

    def _validate_fake_virt_file(self, key):
        file_path = self.get(key)
        # Ensure the value we've got is a file we can access
        try:
            accessible_file(file_path)
        except ValueError as e:
            message = "Error validating key '%s': '%s'" % (key, str(e))
            return 'error', message
        # We do not know if the is_hypervisor key will have been parsed yet
        # TODO Allow for specification of which key requires which parsed values
        try:
            is_hypervisor = self.get('is_hypervisor')
            if isinstance(is_hypervisor, str):
                is_hypervisor = str_to_bool(is_hypervisor)
        except ValueError as e:
            message = "Error validating key '%s': '%s'" % (key, str(e))
            return 'error', message
        try:
            if is_hypervisor is True:
                # Try to read and parse the file as if it were a host guest mapping
                FakeVirt.read_host_guest_mapping_from_file(file_path)
            elif is_hypervisor is False:
                # Try to read and parse the file as if it were a domain list
                FakeVirt.list_domains_from_file(file_path)
        except VirtError as e:
            message = "Error validating key '%s': '%s'" % (key, str(e))
            return 'error', message


class FakeVirt(Virt):
    CONFIG_TYPE = 'fake'

    def __init__(self, logger, config, dest, terminate_event=None,
                 interval=None, oneshot=False, status=False):
        super(FakeVirt, self).__init__(logger, config, dest,
                                       terminate_event=terminate_event,
                                       interval=interval,
                                       oneshot=oneshot,
                                       status=status)
        self.logger = logger
        self.config = config

    @staticmethod
    def _read_data(file_path):
        # TODO: do some checking of the file content
        try:
            with open(file_path, 'r') as f:
                return json.load(f, object_hook=decode)
        except (IOError, ValueError) as e:
            raise VirtError("Can't read fake '%s' virt data: %s" % (file_path, str(e)))

    def isHypervisor(self):
        if self.config["is_hypervisor"] is None:
            return True
        return self.config["is_hypervisor"]

    @classmethod
    def process_guest(cls, guest):
        attributes = guest.get('attributes', {})
        virt_type = attributes.get('virtWhoType', 'fake')
        return Guest(guest['guestId'], virt_type, guest['state'])

    @classmethod
    def process_hypervisor(cls, hypervisor):
        guests = []
        for guest in hypervisor['guestIds']:
            guests.append(cls.process_guest(guest))
        return Hypervisor(hypervisor['hypervisorId']['hypervisorId'],
                          guests,
                          hypervisor.get('name'),
                          hypervisor.get('facts'))

    @classmethod
    def read_host_guest_mapping_from_file(cls, file_path):
        # Used to read the host-guest mapping from a file. Raises VirtError if
        # we are unable to parse the file
        assoc = {'hypervisors': []}
        try:
            for hypervisor in cls._read_data(file_path)['hypervisors']:
                assoc['hypervisors'].append(cls.process_hypervisor(hypervisor))
        except KeyError as e:
            raise VirtError("Fake virt file '%s' is not properly formed: %s" % (file_path, str(e)))
        return assoc

    def getHostGuestMapping(self):
        return self.read_host_guest_mapping_from_file(self.config['file'])

    @classmethod
    def list_domains_from_file(cls, file_path):
        hypervisor = cls._read_data(file_path)['hypervisors'][0]
        if 'hypervisorId' in hypervisor:
            raise VirtError("Fake virt file '%s' is not properly formed: "
                            "uuid key shouldn't be present, try to check is_hypervisor value" %
                            file_path)
        guests = []
        for guest in hypervisor['guestIds']:
            guests.append(cls.process_guest(guest))
        return guests

    def listDomains(self):
        return self.list_domains_from_file(self.config['file'])



#
# Copyright 2019 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA
#
# Refer to the README and COPYING files for full details of the license
#
from __future__ import absolute_import

import atexit
import base64
import os
import tempfile
import urllib3
import yaml


_temp_files = {}


def _create_temp_file_with_content(content):
    if len(_temp_files) == 0:
        atexit.register(_cleanup_temp_files)
    # Because we may change context several times, try to remember files we
    # created and reuse them at a small memory cost.
    content_key = str(content)
    if content_key in _temp_files:
        return _temp_files[content_key]
    _, name = tempfile.mkstemp()
    _temp_files[content_key] = name
    with open(name, 'wb') as fd:
        fd.write(content.encode() if isinstance(content, str) else content)
    return name


def _cleanup_temp_files():
    global _temp_files
    for temp_file in _temp_files.values():
        try:
            os.remove(temp_file)
        except OSError:
            pass
    _temp_files = {}


def _get_kube_config_loader_for_yaml_file(filename, **kwargs):
    with open(filename) as f:
        return KubeConfigLoader(
            config_dict=yaml.load(f),
            config_base_path=os.path.abspath(os.path.dirname(filename)),
            **kwargs)


class Configuration(object):
    pass


class ConfigException(Exception):
    pass


class KubeConfigLoader(object):

    def __init__(self, config_dict, active_context=None,
                 config_base_path="",
                 config_persister=None):
        self._config = ConfigNode('kube-config', config_dict)
        self._current_context = None
        self._user = None
        self._cluster = None
        self.set_active_context(active_context)
        self._config_base_path = config_base_path
        self._config_persister = config_persister

    def set_active_context(self, context_name=None):
        if context_name is None:
            context_name = self._config['current-context']
        self._current_context = self._config['contexts'].get_with_name(
            context_name)
        if (self._current_context['context'].safe_get('user') and
                self._config.safe_get('users')):
            user = self._config['users'].get_with_name(
                self._current_context['context']['user'], safe=True)
            if user:
                self._user = user['user']
            else:
                self._user = None
        else:
            self._user = None
        self._cluster = self._config['clusters'].get_with_name(
            self._current_context['context']['cluster'])['cluster']

    def _load_cluster_info(self):
        if 'server' in self._cluster:
            self.host = self._cluster['server']
            if self.host.startswith("https"):
                self.ssl_ca_cert = FileOrData(
                    self._cluster, 'certificate-authority',
                    file_base_path=self._config_base_path).as_file()
                self.cert_file = FileOrData(
                    self._user, 'client-certificate',
                    file_base_path=self._config_base_path).as_file()
                self.key_file = FileOrData(
                    self._user, 'client-key',
                    file_base_path=self._config_base_path).as_file()
        if 'insecure-skip-tls-verify' in self._cluster:
            self.verify_ssl = not self._cluster['insecure-skip-tls-verify']

    def _load_authentication(self):
        if self._load_user_token():
            return
        self._load_user_pass_token()

    def _load_user_token(self):
        token = FileOrData(
            self._user, 'tokenFile', 'token',
            file_base_path=self._config_base_path,
            base64_file_content=False).as_data()
        if token:
            self.token = "Bearer %s" % token
            return True

    def _load_user_pass_token(self):
        if 'username' in self._user and 'password' in self._user:
            self.token = urllib3.util.make_headers(
                basic_auth=(self._user['username'] + ':' +
                            self._user['password'])).get('authorization')
            return True

    def _set_config(self, client_configuration):
        if 'token' in self.__dict__:
            client_configuration.token = self.token
        # copy these keys directly from self to configuration object
        keys = ['host', 'ssl_ca_cert', 'cert_file', 'key_file', 'verify_ssl']
        for key in keys:
            if key in self.__dict__:
                setattr(client_configuration, key, getattr(self, key))

    def load_and_set(self, client_configuration):
        self._load_authentication()
        self._load_cluster_info()
        self._set_config(client_configuration)

    def list_contexts(self):
        return [context.value for context in self._config['contexts']]

    @property
    def current_context(self):
        return self._current_context.value


class ConfigNode(object):
    """Remembers each config key's path and construct a relevant exception
    message in case of missing keys. The assumption is all access keys are
    present in a well-formed kube-config."""

    def __init__(self, name, value):
        self.name = name
        self.value = value

    def __contains__(self, key):
        return key in self.value

    def __len__(self):
        return len(self.value)

    def safe_get(self, key):
        if (isinstance(self.value, list) and isinstance(key, int) or
                key in self.value):
            return self.value[key]

    def __getitem__(self, key):
        v = self.safe_get(key)
        if not v:
            raise ConfigException(
                'Invalid kube-config file. Expected key %s in %s'
                % (key, self.name))
        if isinstance(v, dict) or isinstance(v, list):
            return ConfigNode('%s/%s' % (self.name, key), v)
        else:
            return v

    def get_with_name(self, name, safe=False):
        if not isinstance(self.value, list):
            raise ConfigException(
                'Invalid kube-config file. Expected %s to be a list'
                % self.name)
        result = None
        for v in self.value:
            if 'name' not in v:
                raise ConfigException(
                    'Invalid kube-config file. '
                    'Expected all values in %s list to have \'name\' key'
                    % self.name)
            if v['name'] == name:
                if result is None:
                    result = v
                else:
                    raise ConfigException(
                        'Invalid kube-config file. '
                        'Expected only one object with name %s in %s list'
                        % (name, self.name))
        if result is not None:
            return ConfigNode('%s[name=%s]' % (self.name, name), result)
        if safe:
            return None
        raise ConfigException(
            'Invalid kube-config file. '
            'Expected object with name %s in %s list' % (name, self.name))


class FileOrData(object):
    """Utility class to read content of obj[%data_key_name] or file's
     content of obj[%file_key_name] and represent it as file or data.
     Note that the data is preferred. The obj[%file_key_name] will be used iff
     obj['%data_key_name'] is not set or empty. Assumption is file content is
     raw data and data field is base64 string. The assumption can be changed
     with base64_file_content flag. If set to False, the content of the file
     will assumed to be base64 and read as is. The default True value will
     result in base64 encode of the file content after read."""

    def __init__(self, obj, file_key_name, data_key_name=None,
                 file_base_path="", base64_file_content=True):
        if not data_key_name:
            data_key_name = file_key_name + "-data"
        self._file = None
        self._data = None
        self._base64_file_content = base64_file_content
        if data_key_name in obj:
            self._data = obj[data_key_name]
        elif file_key_name in obj:
            self._file = os.path.normpath(
                os.path.join(file_base_path, obj[file_key_name]))

    def as_file(self):
        """If obj[%data_key_name] exists, return name of a file with base64
        decoded obj[%data_key_name] content otherwise obj[%file_key_name]."""
        use_data_if_no_file = not self._file and self._data
        if use_data_if_no_file:
            if self._base64_file_content:
                if isinstance(self._data, str):
                    content = self._data.encode()
                else:
                    content = self._data
                self._file = _create_temp_file_with_content(
                    base64.standard_b64decode(content))
            else:
                self._file = _create_temp_file_with_content(self._data)
        if self._file and not os.path.isfile(self._file):
            raise ConfigException("File does not exists: %s" % self._file)
        return self._file

    def as_data(self):
        """If obj[%data_key_name] exists, Return obj[%data_key_name] otherwise
        base64 encoded string of obj[%file_key_name] file content."""
        use_file_if_no_data = not self._data and self._file
        if use_file_if_no_data:
            with open(self._file) as f:
                if self._base64_file_content:
                    self._data = bytes.decode(
                        base64.standard_b64encode(str.encode(f.read())))
                else:
                    self._data = f.read()
        return self._data



from __future__ import print_function

# Default interval for sending list of UUIDs
DefaultInterval = 3600  # One per hour
MinimumSendInterval = 60  # One minute
MinimumJobPollInterval = 15

SAT5 = "satellite"
SAT6 = "sam"



from __future__ import print_function
"""
Test of subscription managers.

Copyright (C) 2014 Radek Novacek <rnovacek@redhat.com>

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""

import os
import shutil
import tempfile
from mock import patch, MagicMock, ANY

from base import TestBase, unittest

from virtwho.config import VirtConfigSection
from virtwho.manager import Manager, ManagerError
from virtwho.virt import Guest, Hypervisor, HostGuestAssociationReport, DomainListReport

import rhsm.config as rhsm_config


xvirt = type("", (), {'CONFIG_TYPE': 'xxx'})()


class TestManager(TestBase):
    """ Test of all available subscription managers. """
    guest1 = Guest('9c927368-e888-43b4-9cdb-91b10431b258', xvirt.CONFIG_TYPE, Guest.STATE_RUNNING)
    guest2 = Guest('d5ffceb5-f79d-41be-a4c1-204f836e144a', xvirt.CONFIG_TYPE, Guest.STATE_SHUTOFF)
    guestInfo = [guest1]
    hypervisor_id = "HYPERVISOR_ID"

    config = VirtConfigSection.from_dict({'type': 'libvirt', 'owner': 'OWNER'}, 'test', None)
    host_guest_report = HostGuestAssociationReport(config, {
        'hypervisors': [
            Hypervisor('9c927368-e888-43b4-9cdb-91b10431b258', []),
            Hypervisor('ad58b739-5288-4cbc-a984-bd771612d670', [guest1, guest2])
        ]
    })
    domain_report = DomainListReport(config, [guest1], hypervisor_id)


class TestSubscriptionManager(TestManager):
    smType = "sam"

    default_config_args = {
        'type': 'libvirt',
        'hypervisor_id': 'uuid',
        'rhsm_username': 'user',
        'rhsm_password': 'password'
    }

    def prepare(self, create_from_file, connection):
        self.options = MagicMock()

        tempdir = tempfile.mkdtemp()
        self.addCleanup(shutil.rmtree, tempdir)

        config_file = os.path.join(tempdir, "config")
        with open(config_file, "w") as f:
            f.write("[rhsm]\nconsumerCertDir=%s\n" % tempdir)

        cert_file = os.path.join(tempdir, "cert.pem")
        with open(cert_file, "w") as f:
            f.write("\n")

        rhsm_config.DEFAULT_CONFIG_PATH = config_file

        create_from_file.return_value.cert_uuid = {'CN': 'Test'}
        connection.return_value = MagicMock()
        connection.return_value.has_capability = MagicMock(return_value=False)
        connection.return_value.getConsumer = MagicMock(return_value={})
        connection.return_value.getOwner = MagicMock(return_value={'key': 'OWNER'})

    @patch("rhsm.connection.UEPConnection")
    @patch("rhsm.certificate.create_from_file")
    def test_sendVirtGuests(self, create_from_file, connection):
        self.prepare(create_from_file, connection)
        config, d = self.create_fake_config('test', **self.default_config_args)
        manager = Manager.from_config(self.logger, config)
        self.domain_report._config = config
        manager.sendVirtGuests(self.domain_report, self.options)
        manager.connection.updateConsumer.assert_called_with(
            ANY,
            guest_uuids=[guest.toDict() for guest in self.guestInfo],
            hypervisor_id=self.hypervisor_id)

    @patch("rhsm.connection.UEPConnection")
    @patch("rhsm.certificate.create_from_file")
    def test_hypervisorCheckIn(self, create_from_file, connection):
        self.prepare(create_from_file, connection)
        config, d = self.create_fake_config('test', **self.default_config_args)
        d['owner'] = 'OWNER'
        manager = Manager.from_config(self.logger, config)
        # TODO additional mocking. Specifically, mock out the host_guest_report and config...
        self.host_guest_report._config = config
        manager.hypervisorCheckIn(self.host_guest_report)
        manager.connection.hypervisorCheckIn.assert_called_with(
            d['owner'],
            '',
            dict(
                (
                    host.hypervisorId,
                    [
                        guest.toDict()
                        for guest in host.guestIds
                    ]
                )
                for host in self.host_guest_report.association['hypervisors']),
            options=ANY)


class TestSatellite(TestManager):
    smType = "satellite"

    default_config_args = {
        'type': 'libvirt',
        'sm_type': 'satellite',
        'hypervisor_id': 'uuid',
        'sat_server': 'localhost',
        'sat_username': 'username',
        'sat_password': 'password',
    }

    @unittest.skip("skip until config section for satellite is implemented")
    def test_sendVirtGuests(self):
        config = VirtConfigSection.from_dict({'type': 'libvirt', 'sat_server': 'localhost'}, 'test', None)
        manager = Manager.from_config(self.logger, config)
        self.assertRaises(ManagerError, manager.sendVirtGuests, self.domain_report)

    @patch("six.moves.xmlrpc_client.ServerProxy")
    def test_hypervisorCheckIn(self, server):
        options = MagicMock()
        server.return_value.registration.new_system_user_pass.return_value = {
            'system_id': '123'
        }
        config, d = self.create_fake_config('test', **self.default_config_args)
        manager = Manager.from_config(self.logger, config)
        self.host_guest_report._config = config
        manager.hypervisorCheckIn(self.host_guest_report, options)
        manager.server_xmlrpc.registration.virt_notify.assert_called_with(ANY, [
            [0, "exists", "system", {"identity": "host", "uuid": "0000000000000000"}],
            [0, "crawl_began", "system", {}],
            [0, "exists", "domain", {
                "memory_size": 0,
                "name": "VM 9c927368-e888-43b4-9cdb-91b10431b258 from libvirt hypervisor ad58b739-5288-4cbc-a984-bd771612d670",
                "state": "running",
                "uuid": "9c927368e88843b49cdb91b10431b258",
                "vcpus": 1,
                "virt_type": "fully_virtualized"
            }],
            [0, "exists", "domain", {
                "memory_size": 0,
                "name": "VM d5ffceb5-f79d-41be-a4c1-204f836e144a from libvirt hypervisor ad58b739-5288-4cbc-a984-bd771612d670",
                "state": "shutoff",
                "uuid": "d5ffceb5f79d41bea4c1204f836e144a",
                "vcpus": 1,
                "virt_type": "fully_virtualized"
            }],
            [0, "crawl_ended", "system", {}]
        ])



from __future__ import print_function
#
# Copyright 2018 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA
#
# Refer to the README and COPYING files for full details of the license
#
from mock import patch, Mock, ANY

from base import TestBase
from threading import Event

from virtwho.virt import Virt, Guest, Hypervisor, StatusReport
from virtwho.virt.kubevirt.kubevirt import KubevirtConfigSection
from virtwho.datastore import Datastore


class TestKubevirt(TestBase):

    @staticmethod
    def create_config(name, wrapper, **kwargs):
        config = KubevirtConfigSection(name, wrapper)
        config.update(**kwargs)
        config.validate()
        return config

    def nodes(self):
        node = {
            'metadata': {
                'name': 'master'
            },
            'status': {
                'nodeInfo': {
                    'machineID': '52c01ad890e84b15a1be4be18bd64ecd',
                    'kubeletVersion': 'v1.9.1+a0ce1bc657'
                },
                'addresses': [
                    {'address': '192.168.122.140',
                     'type': 'InternalIP'},
                    {'address': 'minikube',
                     'type': 'Hostname'}
                ],
                'allocatable' : {
                    'cpu': '2'
                }
            }
        }

        return {'items': [node]}

    def new_nodes(self):
        node = {
            'metadata': {
                'name': 'master'
            },
            'status': {
                'nodeInfo': {
                    'machineID': '52c01ad890e84b15a1be4be18bd64ecd',
                    'kubeletVersion': 'v1.18.0-rc.1'
                },
                'addresses': [
                    {'address': '192.168.122.140',
                     'type': 'InternalIP'},
                    {'address': 'minikube',
                     'type': 'Hostname'}
                ],
                'allocatable' : {
                    'cpu': '7500m'
                }
            }
        }
        return {'items': [node]}

    def vms(self):
        vm = {
            'metadata': {
                'name': 'win-2016',
                'namespace': 'default',
            },
            'spec': {
                'domain': {
                    'devices': {
                        'disks': [
                            {'disk': {'bus': 'virtio'}, 'name': 'containerdisk'}
                        ],
                        'interfaces': [
                            {'bridge': {}, 'name': 'default'}
                        ],
                    },
                    'features': {
                        'acpi': {
                            'enabled': 'true'
                        }
                    },
                    'firmware': {
                        'uuid': 'f83c5f73-5244-4bd1-90cf-02bac2dda608'
                    },
                    'machine': {
                        'type': 'q35'
                    }
                }
            },
            'status': {
                'nodeName': 'master',
                'phase': 'Running',
            }
        }

        return {'items': [vm]}
    
    def pending_vms(self):
        vm = {
            'metadata': {
                'name': 'win-2016',
                'namespace': 'default',
            },
            'spec': {
                'domain': {
                    'devices': {
                        'disks': [
                            {'disk': {'bus': 'virtio'}, 'name': 'containerdisk'}
                        ],
                        'interfaces': [
                            {'bridge': {}, 'name': 'default'}
                        ],
                    },
                    'features': {
                        'acpi': {
                            'enabled': 'true'
                        }
                    },
                    'firmware': {
                        'uuid': 'f83c5f73-5244-4bd1-90cf-02bac2dda608'
                    },
                    'machine': {
                        'type': 'q35'
                    }
                }
            },
            'status': {
                'phase': 'Pending',
            }
        }

        return {'items': [vm]}

    def test_pending_vm(self):
        client = Mock()
        client.get_nodes.return_value = self.nodes()
        client.get_vms.return_value = self.pending_vms()

        config = self.create_config(name='test', wrapper=None, type='kubevirt',
                                    owner='owner', kubeconfig='/etc/hosts')

        with patch.dict('os.environ', {'KUBECONFIG':'/dev/null'}):
            kubevirt = Virt.from_config(self.logger, config, Datastore())

            kubevirt._client = client

            expected_result = Hypervisor(
                hypervisorId='52c01ad890e84b15a1be4be18bd64ecd',
                name='master',
                guestIds=[],
                facts={
                    Hypervisor.CPU_SOCKET_FACT: '2',
                    Hypervisor.HYPERVISOR_TYPE_FACT: 'qemu',
                    Hypervisor.SYSTEM_UUID_FACT: '52c01ad890e84b15a1be4be18bd64ecd',
                    Hypervisor.HYPERVISOR_VERSION_FACT: 'v1.9.1+a0ce1bc657',
                }
            )
            result = kubevirt.getHostGuestMapping()['hypervisors'][0]
            self.assertEqual(expected_result.toDict(), result.toDict())

    def test_getHostGuestMapping(self):
        client = Mock()
        client.get_nodes.return_value = self.nodes()
        client.get_vms.return_value = self.vms()

        config = self.create_config(name='test', wrapper=None, type='kubevirt',
                                    owner='owner', kubeconfig='/etc/hosts')

        with patch.dict('os.environ', {'KUBECONFIG':'/dev/null'}):
            kubevirt = Virt.from_config(self.logger, config, Datastore())

            kubevirt._client = client

            expected_result = Hypervisor(
                hypervisorId='52c01ad890e84b15a1be4be18bd64ecd',
                name='master',
                guestIds=[
                    Guest(
                        'f83c5f73-5244-4bd1-90cf-02bac2dda608',
                        kubevirt.CONFIG_TYPE,
                        Guest.STATE_RUNNING,
                    )
                ],
                facts={
                    Hypervisor.CPU_SOCKET_FACT: '2',
                    Hypervisor.HYPERVISOR_TYPE_FACT: 'qemu',
                    Hypervisor.SYSTEM_UUID_FACT: '52c01ad890e84b15a1be4be18bd64ecd',
                    Hypervisor.HYPERVISOR_VERSION_FACT: 'v1.9.1+a0ce1bc657',
                }
            )
            result = kubevirt.getHostGuestMapping()['hypervisors'][0]
            self.assertEqual(expected_result.toDict(), result.toDict())

    def test_getHostGuestMapping_with_hm(self):
        client = Mock()
        client.get_nodes.return_value = self.nodes()
        client.get_vms.return_value = self.vms()

        config = self.create_config(name='test', wrapper=None, type='kubevirt',
                                    owner='owner', kubeconfig='/etc/hosts',
                                    hypervisor_id='hostname')

        with patch.dict('os.environ', {'KUBECONFIG':'/dev/null'}):
            kubevirt = Virt.from_config(self.logger, config, Datastore())

            kubevirt._client = client

            expected_result = Hypervisor(
                hypervisorId='minikube',
                name='master',
                guestIds=[
                    Guest(
                        'f83c5f73-5244-4bd1-90cf-02bac2dda608',
                        kubevirt.CONFIG_TYPE,
                        Guest.STATE_RUNNING,
                    )
                ],
                facts={
                    Hypervisor.CPU_SOCKET_FACT: '2',
                    Hypervisor.HYPERVISOR_TYPE_FACT: 'qemu',
                    Hypervisor.SYSTEM_UUID_FACT: '52c01ad890e84b15a1be4be18bd64ecd',
                    Hypervisor.HYPERVISOR_VERSION_FACT: 'v1.9.1+a0ce1bc657',
                }
            )
            result = kubevirt.getHostGuestMapping()['hypervisors'][0]
            self.assertEqual(expected_result.toDict(), result.toDict())

    def test_milicpu(self):
        client = Mock()
        client.get_nodes.return_value = self.new_nodes()
        client.get_vms.return_value = self.vms()

        config = self.create_config(name='test', wrapper=None, type='kubevirt',
                                    owner='owner', kubeconfig='/etc/hosts')

        with patch.dict('os.environ', {'KUBECONFIG':'/dev/null'}):
            kubevirt = Virt.from_config(self.logger, config, Datastore())
            kubevirt._client = client

            expected_result = Hypervisor(
                hypervisorId='52c01ad890e84b15a1be4be18bd64ecd',
                name='master',
                guestIds=[
                    Guest(
                        'f83c5f73-5244-4bd1-90cf-02bac2dda608',
                        kubevirt.CONFIG_TYPE,
                        Guest.STATE_RUNNING,
                    )
                ],
                facts={
                    Hypervisor.CPU_SOCKET_FACT: '7',
                    Hypervisor.HYPERVISOR_TYPE_FACT: 'qemu',
                    Hypervisor.SYSTEM_UUID_FACT: '52c01ad890e84b15a1be4be18bd64ecd',
                    Hypervisor.HYPERVISOR_VERSION_FACT: 'v1.18.0-rc.1',
                }
            )
            result = kubevirt.getHostGuestMapping()['hypervisors'][0]
            self.assertEqual(expected_result.toDict(), result.toDict())


    def test_empty_kubeconfig(self):
        config = self.create_config(name='test', wrapper=None, type='kubevirt',
                                    owner='owner')

        kubevirt = Virt.from_config(self.logger, config, Datastore())
        self.assertEqual("~/.kube/config", kubevirt._path)

    @patch("virtwho.virt.kubevirt.config._get_kube_config_loader_for_yaml_file",
           return_value=Mock())
    @patch("virtwho.virt.kubevirt.config.Configuration")
    def test_version_override(self, cfg, _):
        version='v1alpha3'
        cfg.return_value = Config()
        config = self.create_config(name='test', wrapper=None, type='kubevirt',
                                    owner='owner', kubeconfig='/etc/hosts',
                                    kubeversion=version, hypervisor_id='hostname')

        kubevirt = Virt.from_config(self.logger, config, Datastore())
        kubevirt.prepare()
        self.assertEqual(version, kubevirt._version)

    @patch("virtwho.virt.kubevirt.config._get_kube_config_loader_for_yaml_file",
           return_value=Mock())
    @patch("virtwho.virt.kubevirt.config.Configuration")
    def test_insecure(self, cfg, _):
        cfg.return_value = Config()
        config = self.create_config(name='test', wrapper=None, type='kubevirt',
                                    owner='owner', kubeconfig='/etc/hosts',
                                    kubeversion='v1alpha3', hypervisor_id='hostname',
                                    insecure='')      

        kubevirt = Virt.from_config(self.logger, config, Datastore())
        kubevirt.prepare()
        self.assertFalse(kubevirt._insecure)

    @patch("virtwho.virt.kubevirt.config._get_kube_config_loader_for_yaml_file",
           return_value=Mock())
    @patch("virtwho.virt.kubevirt.config.Configuration")
    def test_status(self, cfg, _):
        cfg.return_value = Config()
        client = Mock()
        client.get_nodes.return_value = self.nodes()
        client.get_vms.return_value = self.vms()

        self.config = self.create_config(name='test', wrapper=None, type='kubevirt',
                                    owner='owner', kubeconfig='/etc/hosts',
                                    kubeversion='version', hypervisor_id='hostname')
        self.config['server'] = 'kubeserver'

        with patch.dict('os.environ', {'KUBECONFIG':'/dev/null'}):
            kubevirt = Virt.from_config(self.logger, self.config, Datastore())
            kubevirt.prepare()

            kubevirt._client = client
            kubevirt.status = True
            kubevirt._send_data = Mock()
            self.run_once(kubevirt)

            kubevirt._send_data.assert_called_once_with(data_to_send=ANY)
            self.assertTrue(isinstance(kubevirt._send_data.mock_calls[0].kwargs['data_to_send'], StatusReport))
            self.assertEqual(kubevirt._send_data.mock_calls[0].kwargs['data_to_send'].data['source']['server'],
                              self.config['server'])


    def run_once(self, kubevirt, datastore=None):
        ''' Run kubevirt in oneshot mode '''
        kubevirt._oneshot = True
        if datastore is None:
            datastore = Mock(spec=Datastore())

        kubevirt.dest = datastore
        kubevirt._terminate_event = Event()
        kubevirt._oneshot = True
        kubevirt._interval = 0
        kubevirt._run()


class Config(object):

    ssl_ca_cert = "file"
    cert_file = "file"
    key_file = "file"
    host = "localhost"
    token = "token"



# -*- coding: utf-8 -*-
from __future__ import print_function
"""
Module for reading configuration files

Copyright (C) 2017 Christopher Snyder <csnyder@redhat.com>

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""
import copy
from threading import Lock


class Datastore(object):
    """ This class is a threadsafe datastore"""

    def __init__(self, *args, **kwargs):
        self._datastore = dict()
        self._datastore_lock = Lock()

    def put(self, key, value):
        """
        Stores the value, retrievable by key, in a threadsafe manner in the
        underlying datastore.

        @param key: The unique identifier for this value
        @type  key: str

        @param value: The object to store
        """
        to_store = copy.deepcopy(value)
        with self._datastore_lock:
            self._datastore[key] = to_store

    def get(self, key, default=None):
        """
        Retrieves the value for the given key, in a threadsafe manner from the
        underlying datastore.

        @param key: The unique identifier for this value
        @type  key: str

        @param default: An optional default object to return should the
        underlying datastore not have an item for the given key.

        @raises KeyError: A KeyError is raised when the underlying datastore
        has no item for the given key and a default has not been provided.
        """
        with self._datastore_lock:
            try:
                return self._datastore[key]
            except KeyError:
                if default:
                    return default
                raise



from __future__ import print_function

#
# Module for abstraction of all virtualization backends, part of virt-who
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#

import os
import signal
import sys
import socket
from six.moves.queue import Empty
from shutil import rmtree
from multiprocessing import Process, Queue
from mock import patch

import json
from tempfile import TemporaryFile, mkdtemp

from fake_sam import FakeSam

import virtwho
import virtwho.parser
import virtwho.log
import virtwho.config
import virtwho.lock

from unittest import TestCase


class TestBase(TestCase):
    """
    This is a class that provides tests for virt backend. The fake backend
    must provide exactly the expected data.
    """

    # Process representing fake candlepin server
    sam = None
    # Process representing fake virt backend server
    server = None
    # Queue used for multiprocess communication
    queue = None
    # Arguments shared for all tests
    arguments = None
    # Virt type
    virt = None
    # Temporary directory used for storing virt-who configuration file
    config_dir = None

    def __init__(self, *args, **kwargs):
        super(TestBase, self).__init__(*args, **kwargs)

    @classmethod
    def setUpClass(cls):
        TestCase.setUpClass()
        virtwho.config.VW_CONF_DIR = '/this/does/not/exist'
        virtwho.config.VW_GENERAL_CONF_PATH = '/this/does/not/exist.conf'
        cls.queue = Queue()
        cls.sam = FakeSam(cls.queue)
        cls.sam.start()

    @classmethod
    def tearDownClass(cls):
        cls.sam.terminate()
        cls.sam.join()
        TestCase.tearDownClass()

    def setUp(self):
        # Clear the queue
        while True:
            try:
                self.queue.get(block=False)
            except Empty:
                break
        self.server.data_version = 0

        # Logger patching
        self.tmp_dir = mkdtemp()
        logger_patcher = patch.multiple('virtwho.log.Logger', _log_dir=self.tmp_dir,
                                        _stream_handler=None, _queue_logger=None)
        logger_patcher.start()
        self.addCleanup(logger_patcher.stop)
        self.addCleanup(rmtree, self.tmp_dir)

        log_patcher = patch.multiple('virtwho.log', DEFAULT_LOG_DIR=self.tmp_dir)
        log_patcher.start()
        self.addCleanup(log_patcher.stop)

        rhsm_log_patcher = patch('rhsm.connection.log')
        rhsm_log_patcher.start()
        self.addCleanup(rhsm_log_patcher.stop)

        # Reduce minimum send interval to allow for faster test completion
        minimum_patcher = patch('virtwho.config.MinimumSendInterval', 2)
        minimum_patcher.start()
        self.addCleanup(minimum_patcher.stop)

        # Mock PIDFILE (so we can run tests as an unprivledged user)
        pid_file_name = self.tmp_dir + os.path.sep + 'virt-who.pid'
        pid_file_patcher = patch('virtwho.main.PIDFILE', pid_file_name)
        pid_file_patcher.start()
        self.addCleanup(pid_file_patcher.stop)

        # Mock status files (so we can run tests as an unprivledged user)
        status_pid_file_name = self.tmp_dir + os.path.sep +'virt-who-status.pid'
        status_pid_file_patcher = patch('virtwho.executor.STATUS_LOCK', status_pid_file_name)
        status_pid_file_patcher.start()
        self.addCleanup(status_pid_file_patcher.stop)

        status_pid_file_patcher = patch('virtwho.virt.virt.STATUS_LOCK', status_pid_file_name)
        status_pid_file_patcher.start()
        self.addCleanup(status_pid_file_patcher.stop)

        status_data_dir_patcher = patch('virtwho.executor.STATUS_DATA_DIR', self.tmp_dir)
        status_data_dir_patcher.start()
        self.addCleanup(status_data_dir_patcher.stop)

        status_file_name = self.tmp_dir + os.path.sep +'run_data.json'
        status_data_file_patcher = patch('virtwho.executor.STATUS_DATA', status_file_name)
        status_data_file_patcher.start()
        self.addCleanup(status_data_file_patcher.stop)

        status_data_file_patcher = patch('virtwho.virt.virt.STATUS_DATA', status_file_name)
        status_data_file_patcher.start()
        self.addCleanup(status_data_file_patcher.stop)

    def tearDown(self):
        self.process.terminate()
        self.process.join()

    def run_virtwho(self, args, grab_stdout=False, background=False):
        """
        Execute virt-who process with given arguments.

        `grab_stdout` argument will take stdout of the virt-who process
        and will return it as second part of the tuple.

        If `background` is set to True, virt-who will be started on background
        and this method will return immediately.

        Returns tuple (status, stdout), where status is return code of the
        virt-who process (or None if `background` is True) and stdout is
        stdout from the process (or None if `grab_stdout` is False).
        """

        old_stdout = None
        if grab_stdout:
            old_stdout = sys.stdout
            sys.stdout = TemporaryFile(mode='w+')
        sys.argv = ["virt-who"] + args
        code = None
        data = None
        socket.setdefaulttimeout(1)
        from virtwho.__main__ import main as main_main
        self.process = Process(target=main_main)
        self.process.start()

        if not background:
            self.process.join()
            code = self.process.exitcode

        if grab_stdout:
            sys.stdout.seek(0)
            data = sys.stdout.read()
            sys.stdout.close()
            sys.stdout = old_stdout

        return code, data

    def stop_virtwho(self):
        self.process.terminate()
        self.process.join()

    def check_assoc(self, reported, expected):
        diff = set(reported.keys()) - set(expected.keys())
        self.assertEqual(len(diff), 0, "Hosts %s reported but not expected" % ",".join(diff))
        diff = set(expected.keys()) - set(reported.keys())
        self.assertEqual(len(diff), 0, "Hosts %s expected but not reported" % ",".join(diff))
        for host, excepted_guests in list(expected.items()):
            reported_guests = reported[host]
            expected_guests_uuids = [guest['guestId'] for guest in excepted_guests]
            reported_guests_uuids = [guest['guestId'] for guest in reported_guests]
            diff = set(reported_guests_uuids) - set(expected_guests_uuids)
            self.assertEqual(len(diff), 0, "Guests %s on host %s reported but not expected" % (",".join(diff), host))
            diff = set(expected_guests_uuids) - set(reported_guests_uuids)
            self.assertEqual(len(diff), 0, "Guests %s on host %s expected but not reported" % (",".join(diff), host))
            for excepted_guest in excepted_guests:
                reported_guest = [guest for guest in reported_guests if guest['guestId'] == excepted_guest['guestId']][0]
                self.assertEqual(excepted_guest, reported_guest, "Guest attributes differs")

    def check_assoc_initial(self, assoc):
        self.check_assoc(assoc, {
            'a2c85a15-9b53-493d-9731-8b5cccdd8951': [],
            '4172853d-e72a-493a-883b-8761f5daa5eb': [],
            '5627a268-f036-4f5d-b9a3-0183ec736913': [
                {
                    'guestId': '9844af5d-101b-40ea-a125-8bf1a02f888b',
                    'attributes': {
                        'active': 1,
                        'virtWhoType': self.virt,
                    },
                    'state': 1
                }, {
                    'guestId': '640bb2fe-fa3b-48cb-89d0-193c13b15663',
                    'attributes': {
                        'active': 1,
                        'virtWhoType': self.virt,
                    },
                    'state': 1
                }, {
                    'guestId': 'c0667b9d-64e1-480c-8b82-c1b1c06614e7',
                    'attributes': {
                        'active': 1,
                        'virtWhoType': self.virt,
                    },
                    'state': 1
                }
            ]
        })

    def check_assoc_updated(self, assoc):
        self.check_assoc(assoc, {
            'a2c85a15-9b53-493d-9731-8b5cccdd8951': [
            ],
            '4172853d-e72a-493a-883b-8761f5daa5eb': [
                {
                    'guestId': '9844af5d-101b-40ea-a125-8bf1a02f888b',
                    'attributes': {
                        'active': 1,
                        'virtWhoType': self.virt,
                    },
                    'state': 1
                }
            ],
            '5627a268-f036-4f5d-b9a3-0183ec736913': [
                {
                    'guestId': '640bb2fe-fa3b-48cb-89d0-193c13b15663',
                    'attributes': {
                        'active': 1,
                        'virtWhoType': self.virt,
                    },
                    'state': 1
                }, {
                    'guestId': 'c0667b9d-64e1-480c-8b82-c1b1c06614e7',
                    'attributes': {
                        'active': 0,
                        'virtWhoType': self.virt,
                    },
                    'state': 5
                }
            ]
        })

    def wait_for_assoc(self, timeout=4):
        try:
            return self.queue.get(timeout=timeout)
        except Empty:
            raise AssertionError("Association was not obtained in %d seconds" % timeout)

    def test_basic(self):
        code, _ = self.run_virtwho(self.arguments + ['-o', '--debug'])
        self.assertEqual(code, 0, "virt-who exited with wrong error code: %s" % code)
        assoc = self.wait_for_assoc()
        self.check_assoc_initial(assoc)

    def test_print(self):
        args = self.arguments + ['-p', '--debug']
        code, out = self.run_virtwho(args, grab_stdout=True)
        self.assertEqual(code, 0, "virt-who exited with wrong error code: %s" % code)
        returned = json.loads(out)
        # Test facts
        for hypervisor in returned['hypervisors']:
            # Check that required facts were obtained from virt backend
            self.assertIn('facts', hypervisor)
            self.assertIn('hypervisor.type', hypervisor['facts'])
            self.assertIn('hypervisor.cluster', hypervisor['facts'])
            if hypervisor['facts']['hypervisor.type'] == 'vmware':
                self.assertTrue(hypervisor['facts']['hypervisor.cluster'].startswith('ha-cluster-1'))
            else:
                self.assertTrue(hypervisor['facts']['hypervisor.cluster'].startswith('ha-compute-res'))
        assoc = dict((host['hypervisorId']['hypervisorId'], host['guestIds']) for host in returned['hypervisors'])
        self.check_assoc(assoc, {
            'a2c85a15-9b53-493d-9731-8b5cccdd8951': [],
            '4172853d-e72a-493a-883b-8761f5daa5eb': [],
            '5627a268-f036-4f5d-b9a3-0183ec736913': [
                {
                    'guestId': '9844af5d-101b-40ea-a125-8bf1a02f888b',
                    'attributes': {
                        'virtWhoType': self.virt,
                        'active': 1,
                    },
                    'state': 1
                }, {
                    'guestId': '640bb2fe-fa3b-48cb-89d0-193c13b15663',
                    'attributes': {
                        'virtWhoType': self.virt,
                        'active': 1,
                    },
                    'state': 1
                }, {
                    'guestId': 'c0667b9d-64e1-480c-8b82-c1b1c06614e7',
                    'attributes': {
                        'virtWhoType': self.virt,
                        'active': 1,
                    },
                    'state': 1
                }
            ]
        })

    def test_normal(self):
        self.run_virtwho(['-i', '2', '-d'] + self.arguments, background=True)
        self.addCleanup(self.stop_virtwho)

        assoc = self.wait_for_assoc()
        self.check_assoc_initial(assoc)

        self.server.data_version = 1

        assoc = self.wait_for_assoc(5)
        self.check_assoc_updated(assoc)

    def test_exit_on_SIGTERM(self):
        """
        This test shows that virt-who exits cleanly in response to the
        SIGTERM signal
        """
        self.run_virtwho(['-i', '2', '-d'] + self.arguments, background=True)
        self.addCleanup(self.stop_virtwho)
        self.assertEqual(self.process.is_alive(), True)
        os.kill(self.process.pid, signal.SIGTERM)
        self.process.join(timeout=3)
        self.assertEqual(self.process.is_alive(), False)

    def test_reload_on_SIGHUP(self):
        """
        This tests that the rhsm.conf is read once again when the process
        receives the reload signal
        """
        rhsm_conf_path = os.path.join(self.sam.tempdir, 'rhsm.conf')
        with open(rhsm_conf_path, 'r') as f:
            good_rhsm_conf = ''.join(line for line in f.readlines())
        bad_conf = """
[server]
hostname = BADHOSTNAME
prefix = /nogood
port = {port}1337
insecure = 1
proxy_hostname =
""".format(port=self.sam.port)
        with open(os.path.join(self.sam.tempdir, 'rhsm.conf'), 'w') as \
                rhsm_conf_file:
            rhsm_conf_file.write(bad_conf)
            rhsm_conf_file.flush()
        self.run_virtwho(['-i', '2', '-d'] + self.arguments, background=True)
        self.addCleanup(self.stop_virtwho)
        self.assertEqual(self.process.is_alive(), True)

        # We expect the queue to be empty until we the appropriate
        # configuration file is added
        self.assertRaises(AssertionError, self.wait_for_assoc, 1)

        # Update the configuration file with the good one that came from
        # fake_sam
        with open(os.path.join(self.sam.tempdir, 'rhsm.conf'), 'w') as \
                rhsm_conf_file:
            rhsm_conf_file.write(good_rhsm_conf)
        os.kill(self.process.pid, signal.SIGHUP)
        self.wait_for_assoc(4)
        self.assertEqual(self.process.is_alive(), True)


# -*- coding: utf-8 -*-
from __future__ import print_function

#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#

"""
Test validating of configuration values.
"""

from base import TestBase
from mock import Mock
import six
from virtwho.config import ConfigSection, ValidationState

MY_SECTION_NAME = 'my_section'

MY_SECTION_VALUES = {
    'my_str': 'foo',
    'my_bool': False,
    'my_list': ['cat', 'dog', 'frog'],
    'must_have': 'bar'
}


class MyConfigSection(ConfigSection):
    """
    Example of ConfigSection subclass used for unit testing
    """

    def __init__(self, *args, **kwargs):
        super(MyConfigSection, self).__init__(*args, **kwargs)
        self.add_key(key='my_str', validation_method=self._validate_non_empty_string,
                     default='bar')
        self.add_key(key='my_bool', validation_method=self._validate_str_to_bool, default=True)
        self.add_key(key='must_have', validation_method=lambda *args: None,
                     required=True)
        self.add_key(key="my_list", validation_method=self._validate_list, default=[])


class TestConfigSection(TestBase):
    """
    Class used for testing of ConfigSection
    """
    __marker = object()

    def __init__(self, *args, **kwargs):
        super(TestConfigSection, self).__init__(*args, **kwargs)
        self.my_config = None

    def setUp(self):
        self.init_config_section()

    def init_config_section(self):
        """
        This method is executed before each unit test
        """
        self.my_config = MyConfigSection(MY_SECTION_NAME, None)
        # Fill config with some values
        for key, value in MY_SECTION_VALUES.items():
            self.my_config[key] = value

    def test_validate_simple(self):
        """
        Test validation, when all values are set properly (in setUp)
        """
        self.assertEqual(self.my_config.state, ValidationState.NEEDS_VALIDATION)
        result = self.my_config.validate()
        self.assertEqual(self.my_config.name, MY_SECTION_NAME)
        self.assertEqual(result, [])
        self.assertEqual(dict(self.my_config), MY_SECTION_VALUES)
        self.assertEqual(self.my_config.state, ValidationState.VALID)
        self.assertEqual(self.my_config._unvalidated_keys, set())
        self.assertEqual(self.my_config._invalid_keys, set())

    def test_validate_required_option(self):
        """
        Test validation, when required option is missing (with no default)
        """
        del self.my_config['must_have']
        result = self.my_config.validate()
        expected_results = [
            (
                'error',
                'Required option: "must_have" not set.'
            )
        ]
        self.assertEqual(result, expected_results)

    def test_validate_no_options(self):
        """
        Test validation, when there are no options
        """
        self.my_config = MyConfigSection(MY_SECTION_NAME, None)
        result = self.my_config.validate()
        expected_results = [
            (
                'warning',
                'No values provided in: my_section'
            ),
            (
                'error',
                'Required option: "must_have" not set.'
            )
        ]
        self.assertEqual(result, expected_results)
        self.assertEqual(dict(self.my_config), dict(self.my_config.defaults))

    def test_validate_unsupported_option(self):
        """
        Test validation, when there is one more option
        """
        self.my_config['unsupported_opt'] = 'foo'
        result = self.my_config.validate()
        expected_result = [
            (
                'warning',
                'Ignoring unknown configuration option "unsupported_opt"'
            )
        ]
        self.assertEqual(result, expected_result)
        self.assertEqual(self.my_config.state, ValidationState.VALID)

    def test_validate_missing_options_with_default_value(self):
        """
        Test validation, when there is mission option that has defined default value
        """
        self.my_config = MyConfigSection(MY_SECTION_NAME, None)
        # Add required option and one with default value
        self.my_config['must_have'] = 'foo'
        # Add one option with default value, but not the other options (my_list and my_str)
        self.my_config['my_bool'] = True
        result = self.my_config.validate()
        expected_result = [
            ('debug', 'Value for "my_list" not set, using default: []'),
            ('debug', 'Value for "my_str" not set, using default: bar')
        ]
        # We do not particularly care about the ordering here, just that the lists contain the same
        # elements.
        six.assertCountEqual(self, result, expected_result)
        self.assertEqual(self.my_config['my_str'], 'bar')
        self.assertEqual(self.my_config.state, ValidationState.VALID)

    def test_validate_empty_string_with_default_value(self):
        """
        Test validation, when there is empty string with defined default value
        """
        self.my_config = MyConfigSection(MY_SECTION_NAME, None)
        # Add required option and one with default value
        self.my_config['must_have'] = 'foo'
        # Add one option with default value, but not the other options (my_list and my_str)
        self.my_config['my_bool'] = True
        self.my_config['my_list'] = ['foo', 'bar']
        self.my_config['my_str'] = ''
        result = self.my_config.validate()
        expected_result = [
            ('warning', '"my_str" cannot be empty, using default: "bar"')
        ]
        self.assertEqual(result, expected_result)
        self.assertEqual(self.my_config['my_str'], 'bar')
        self.assertEqual(self.my_config.state, ValidationState.VALID)

    def test_validate_wrong_string_with_default_value(self):
        """
        Test validation, when there is empty string with defined default value
        """
        self.my_config = MyConfigSection(MY_SECTION_NAME, None)
        # Add required option and one with default value
        self.my_config['must_have'] = 'foo'
        # Add one option with default value, but not the other options (my_list and my_str)
        self.my_config['my_bool'] = True
        self.my_config['my_list'] = ['foo', 'bar']
        self.my_config['my_str'] = ['foo', 'Foo']
        result = self.my_config.validate()
        expected_result = [
            ('warning', '"my_str" is not set to a valid string, using default: "bar"')
        ]
        self.assertEqual(result, expected_result)
        self.assertEqual(self.my_config['my_str'], 'bar')
        self.assertEqual(self.my_config.state, ValidationState.VALID)

    def test_validate_wrong_bool_option_with_default_value(self):
        """
        Test validation, when there is wrong bool option that has defined default value
        """
        self.my_config['my_bool'] = 'nein'
        result = self.my_config.validate()
        expected_result = [
            (
                'warning',
                'my_bool must be a valid boolean, using default. See man virt-who-config for more info'
            ),
        ]
        self.assertEqual(result, expected_result)
        self.assertEqual(self.my_config['my_bool'], True)
        self.assertEqual(self.my_config.state, ValidationState.VALID)

    def test_update_values(self):
        """
        Test updating values
        """
        my_config = MyConfigSection(MY_SECTION_NAME, None)
        my_config.update(**MY_SECTION_VALUES)
        for key, value in my_config.items():
            self.assertEqual(value, MY_SECTION_VALUES[key])

    # ---------- ADD_KEY AND RELATED VALIDATION TESTS ----------------------------------
    # Note about the below tests:
    # Some of the functionality of ConfigSection.validate (and add_key) is dependant on the
    # values returned by the validation_method given. Also, at present, some of the
    # validation_methods used in subclasses modify the values of the containing class (by
    # accessing _values etc). In my opinion this should stop.
    # Current behaviour for a validation_method is:
    #    take in (self, key)
    #    return None (if things are ok) OR a tuple (or list of tuples) of (log_level,
    #        presumably_error_message)
    #
    # Ideal future behaviour:
    #    Take in all params as kwargs only
    #        the ConfigSection can pass the current values as 'values'
    #    return a KeyValidationResult
    #        which has a known set of attributes that are set (including state to indicate what
    #        to do with the particular value)

    def test_add_key_no_validation_method(self):
        # Tests that given a key and nothing else, add_key fails
        config = ConfigSection('test', None)
        self.assertRaises(AttributeError, config.add_key, key='test')

    def test_add_key_destination(self):
        # Tests add_key with a key, destination (and validation_method) given
        config = ConfigSection('test', None)
        mock_validate_method = Mock()
        # Returning nothing from a validation_method means that the key's value is valid
        # TODO Don't be magical with the return values
        mock_validate_method.return_value = None
        # The below should not blow up
        config.add_key(key='test_key', validation_method=mock_validate_method,
                       destination='test_key_dest')
        test_values = {'test_key': 'test_key_value'}
        config.update(**test_values)
        config.validate()
        # The validate_method should have been called with the ConfigSection and the key
        mock_validate_method.assert_called_once_with('test_key')
        self.assertNotIn('test_key', config, "The config contains the original test_key, "
                                             "should be removed post validate")
        self.assertIn('test_key_dest', config, "The config does not contain a value at the "
                                               "destination specified in the add_key call")
        self.assertEqual(config['test_key_dest'], test_values['test_key'])
        self.assertEqual(config.state, ValidationState.VALID)

    def test_add_key_required(self):
        # Tests add_key with a key, destination (and validation_method) given
        config = ConfigSection('test', None)
        mock_validate_method = Mock()
        # Returning nothing from a validation_method means that the key's value is valid
        # TODO Don't be magical with the return values
        mock_validate_method.return_value = None
        # The below should not blow up
        config.add_key(key='test_key', validation_method=mock_validate_method, required=True)
        test_values = {'test_key': 'test_key_value'}
        config.update(**test_values)
        config.validate()
        # The validate_method should have been called with the key (bound methods would also have
        #  been passed the config section as the first arg)
        # FIXME This seems broken, we should let folks add functions that are not bound to the
        # class instance as validation_methods
        mock_validate_method.assert_called_once_with('test_key')
        self.assertIn('test_key', config, "The config does not contain the test_key, deleted by "
                                          "accident?")
        self.assertEqual(config['test_key'], test_values['test_key'])
        self.assertEqual(config.state, ValidationState.VALID)

    def test_add_key_required_missing(self):
        # Tests add_key with a key, destination (and validation_method) given
        config = ConfigSection('test', None)
        mock_validate_method = Mock()
        # Returning nothing from a validation_method means that the key's value is valid
        # TODO Don't be magical with the return values
        mock_validate_method.return_value = None
        # The below should not blow up
        config.add_key(key='test_key', validation_method=mock_validate_method, required=True)
        test_values = {'test_key_different': 'test_key_different_value'}
        config.update(**test_values)
        config.validate()
        # The validate_method should not have been called (the associated key was not included)
        mock_validate_method.assert_not_called()
        self.assertNotIn('test_key', config, "The 'test_key' was added mistakenly somehow")
        self.assertEqual(config.state, ValidationState.INVALID)

    def test_add_key_default_no_value_provided(self):
        # Tests that a key added with a default uses that default, when not provided a value for
        # the same.
        config = ConfigSection('test', None)
        mock_validate_method = Mock()
        # Returning nothing from a validation_method means that the key's value is valid
        # TODO Don't be magical with the return values
        mock_validate_method.return_value = None
        config.add_key(key='test_key', validation_method=mock_validate_method,
                       default="test_key_default")
        config.validate()
        # We only run the validation_methods on keys that are set (not on the defaults)
        mock_validate_method.assert_not_called()
        self.assertIn('test_key', config, "The default value was not added (check "
                                          "reset_to_defaults?)")
        # The value returned by the config should be that set as the default
        self.assertEqual(config['test_key'], "test_key_default")

    def test_add_key_default_valid_value_provided(self):
        # Tests that a key added with a default DOES NOT USE that default, when the provided
        # value is valid
        config = ConfigSection('test', None)
        mock_validate_method = Mock()
        # Returning nothing from a validation_method means that the key's value is valid
        # TODO Don't be magical with the return values
        mock_validate_method.return_value = None
        config.add_key(key='test_key', validation_method=mock_validate_method,
                       default="test_key_default")
        values = {'test_key': 'test_key_value'}
        config.update(values)
        config.validate()
        mock_validate_method.assert_called_once_with('test_key')
        self.assertIn('test_key', config)
        # The value returned by the config should still be the valid one given it
        self.assertEqual(config['test_key'], "test_key_value")
        self.assertEqual(config.state, ValidationState.VALID)

    def test_add_key_default_invalid_value_provided(self):
        # Tests that a key added with a default DOES NOT USE that default
        # when an invalid value is given as the value to be validated
        config = ConfigSection('test', None)
        mock_validate_method = Mock()
        # Returning something (normally a tuple of (log_level, error_message)) from a validation
        # method with the log_level of 'error' means that the key is invalid
        # TODO Create a more meaningful way of validating values (and presenting messages about
        # validation to the user)
        mock_validate_method.return_value = ('error', 'THIS IS A BAD WAY TO GO...')
        config.add_key(key='test_key', validation_method=mock_validate_method,
                       default="test_key_default")
        values = {'test_key': 'PRETEND_BAD_VALUE'}
        config.update(values)
        config.validate()
        mock_validate_method.assert_called_once_with('test_key')
        self.assertIn('test_key', config)
        # The value that the invalid config holds on to should be the one that was provided,
        # NOT the default (which we assume is good to use)
        self.assertEqual(config['test_key'], values['test_key'])
        self.assertEqual(config.state, ValidationState.INVALID)

    def test_add_key_default_required_no_value_provided(self):
        # Tests that the config is still considered valid when a required key is not provided (
        # but has a default)
        config = ConfigSection('test', None)
        mock_validate_method = Mock()
        mock_validate_method.return_value = None
        config.add_key(key='test_key', validation_method=mock_validate_method, required=True,
                       default='test_key_default')
        config.validate()
        # Not given a value to chew on, don't try to chew on it
        mock_validate_method.assert_not_called()
        self.assertIn('test_key', config, "The config is missing the required, defaulted value ("
                                          "check reset_to_defaults?)")
        self.assertEqual(config['test_key'], 'test_key_default')
        self.assertEqual(config.state, ValidationState.VALID, "Check _unvalidated_keys, "
                                                              "_invalid_keys, _missing_required, "
                                                              "and _update_state()")

    def test_add_key_destination_required_valid_value(self):
        # Tests add_key with a key, destination (and validation_method) given that the key is
        # required
        config = ConfigSection('test', None)
        mock_validate_method = Mock()
        # Returning nothing from a validation_method means that the key's value is valid
        # TODO Don't be magical with the return values
        mock_validate_method.return_value = None
        # The below should not blow up
        config.add_key(key='test_key', validation_method=mock_validate_method,
                       destination='test_key_dest', required=True)
        test_values = {'test_key': 'test_key_value'}
        config.update(**test_values)
        config.validate()
        mock_validate_method.assert_called_once_with('test_key')
        self.assertNotIn('test_key', config, "The config contains the original test_key, "
                                             "should be removed post validate")
        self.assertIn('test_key_dest', config, "The config does not contain a value at the "
                                               "destination specified in the add_key call")
        self.assertEqual(config['test_key_dest'], test_values['test_key'])
        self.assertEqual(config.state, ValidationState.VALID)

    def test_add_key_destination_invalid_value(self):
        # Tests add_key with a key, destination (and validation_method) given that the key is
        # required
        config = ConfigSection('test', None)
        mock_validate_method = Mock()
        # TODO Don't be magical with the return values
        mock_validate_method.return_value = ('error', 'THIS IS A TERRIBLE IDEA...')
        # The below should not blow up
        config.add_key(key='test_key', validation_method=mock_validate_method,
                       destination='test_key_dest')
        test_values = {'test_key': 'test_key_value'}
        config.update(**test_values)
        config.validate()
        mock_validate_method.assert_called_once_with('test_key')
        self.assertNotIn('test_key_dest', config, "The config has a value for the destination, "
                                                  "despite the value being invalid")
        self.assertEqual(config.state, ValidationState.INVALID)

    def test_add_key_validation_method(self):
        # The most minimal use of add_key
        # Verifies that the standard validate method does what is dictated by add_key in
        # the minimal case.
        config = ConfigSection('test', None)
        mock_validate_method = Mock()
        # Returning nothing from a validation_method means that the key's value is valid
        # TODO Don't be magical with the return values
        mock_validate_method.return_value = None
        # The below should not blow up
        config.add_key(key='test_key', validation_method=mock_validate_method)
        test_values = {'test_key': 'test_key_value'}
        config.update(**test_values)
        config.validate()
        # The validate_method should have been called with the ConfigSection and the key
        mock_validate_method.assert_called_once_with('test_key')
        self.assertIn('test_key', config, "The config does not contain the test_key, deleted by "
                                          "accident?")
        self.assertEqual(config['test_key'], test_values['test_key'])
        self.assertEqual(config.state, ValidationState.VALID)

    def _run_add_key_test(self, key, input_value=__marker, expected_value=__marker, add_key_kwargs=None,
                          expected_in=True,
                          expected_state=ValidationState.VALID,
                          validation_return_val=None,
                          expected_present_in_messages=True):
        if not add_key_kwargs:
            add_key_kwargs = {}

        # Provide mock validation_method if one is not given
        if 'validation_method' not in add_key_kwargs:
            add_key_kwargs['validation_method'] = Mock()
            add_key_kwargs['validation_method'].return_value = validation_return_val

        config = ConfigSection('test', None)

        config.add_key(key, **add_key_kwargs)

        # Only update the values in the config section if we've been given an input value
        if input_value is not self.__marker:
            test_values = {key: input_value}
            config.update(**test_values)

        messages = config.validate()

        # Assert the mock was called, if it was a mock (and we passed in an input_value
        if input_value is not self.__marker and \
                hasattr(add_key_kwargs['validation_method'], 'assert_called_once_with'):
            add_key_kwargs['validation_method'].assert_called_once_with(key)

        if expected_in:
            self.assertIn(key, config, "The config does not contain the expected key")

        if expected_value is not self.__marker:
            self.assertEqual(config[key], expected_value, 'Expected key "%s" to be "%s" found "%s"' % (key, expected_value, config[key]))

        if expected_present_in_messages:
            self.assertTrue(any(key in message[1] for message in messages), 'Expected at least one mention of key "%s" in validation messages' % key)
        else:
            bad_messages = []
            for message in messages:
                if key in message[1]:
                    bad_messages.append(message)
            if bad_messages:
                self.fail("Expected no mention of key '%s' in messages found: %s" % (key, bad_messages))

        self.assertEqual(config.state, expected_state)
        return config, messages

    def test_add_key_restricted_with_default(self):
        # Tests that a key added as restricted is not mentioned in validation output
        # Restricted keys are meant to be included in the validation process but are not mentioned
        # when missing with a default.
        add_key_kwargs = {
            'default': "default",
            'restricted': True,
        }
        self._run_add_key_test('test_key', add_key_kwargs=add_key_kwargs,
                               expected_value=add_key_kwargs['default'],
                               expected_present_in_messages=False
                               )
        # Now with a value provided
        self._run_add_key_test('test_key', add_key_kwargs=add_key_kwargs,
                               expected_in=True, input_value='input',
                               expected_value='input',
                               expected_present_in_messages=False
                               )

    def test_add_key_restricted_and_required(self):
        # Tests that a key added as restricted is not mentioned in validation output
        # Restricted keys are meant to be included in the validation process but are not mentioned
        # when missing with a default.
        no_default = dict(
                restricted=True,
                required=True,
        )
        with_default = dict(
                restricted=True,
                required=True,
                default="default"
        )
        # Tuples of args
        cases = [
            # No input_value given, and no default
            dict(
                add_key_kwargs=no_default,
                expected_in=False,
                expected_state=ValidationState.INVALID,
                expected_present_in_messages=True,
            ),
            # A value given, expect the same value, no default
            dict(
                add_key_kwargs=no_default,
                input_value='value',
                expected_value='value',
                expected_state=ValidationState.VALID,
                expected_present_in_messages=False,
            ),
            # Default given, no input value, expect default value
            dict(
                add_key_kwargs=with_default,
                expected_value=with_default['default'],
                expected_state=ValidationState.VALID,
                expected_present_in_messages=False,
            ),
            # Default given, with an input, expect input value
            dict(
                add_key_kwargs=with_default,
                input_value='value',
                expected_value='value',
                expected_state=ValidationState.VALID,
                expected_present_in_messages=False,
            ),
        ]

        for index, case in enumerate(cases):
            try:
                self._run_add_key_test('test_key', **case)
            except AssertionError:
                print("Assertion error during case #%s" % index)
                raise
